/* Copyright (C) 2007 Christian Dywan <christian at twotoasts dot de>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. */

#ifndef __DEBUG_H__
#define __DEBUG_H__ 1

#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
#endif
#if defined(HAVE_UNISTD_H) && SOKOKE_DEBUG > 0
 #include <unistd.h>
 #include <signal.h>
#endif
#include <glib.h>

#if SOKOKE_DEBUG > 0

 #ifndef __func__
  #define __func__ __FUNCTION__
 #endif

 void sokoke_debug(const gchar* sFile, const gchar* sFunc, gint iLine)
 {
  static gchar sDebug[256];
  g_snprintf(sDebug, 255, "%s %s %d", sFile, sFunc, iLine);
  #if SOKOKE_DEBUG < 2
  if(!iLine)
  #endif
   g_print("%s\n", sDebug);
 }

 #define UNIMPLEMENTED g_print(" * Unimplemented: %s\n", __func__);
 #define DEBUG_FUNC sokoke_debug(__FILE__, __func__, __LINE__);

 #if HAVE_UNISTD_H && HAVE_STDLIB_H
 RETSIGTYPE signal_crash_handler(int signal)
 {
  g_print("Terminated: %s\n", g_strsignal(signal));
  sokoke_debug(NULL, NULL, 0); exit(1);
 }
 #endif

#else

 #define UNIMPLEMENTED
 #define DEBUG_FUNC ;

#endif

#if HAVE_UNISTD_H && HAVE_STDLIB_H && SOKOKE_DEBUG > 0
 #define INIT_CRASHSAFE \
  signal(SIGSEGV, signal_crash_handler); \
  signal(SIGABRT, signal_crash_handler);
#else
 #define INIT_CRASHSAFE ;
#endif

#endif /* !__DEBUG_H__ */
