/* Copyright (C) 2007 Christian Dywan <christian at twotoasts dot de>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. */

#include "document.h"

void on_document_location_changed(GtkWidget* document
#ifdef MIDORI_WEBKIT
 , const gchar* sUri
#endif
, CBrowser* browser)
{
    DEBUG_FUNC
    #ifndef MIDORI_WEBKIT
    const gchar* sUri = document_get_uri(document);
    #endif
    browser->sUri = g_strdup(sUri);
    if(document == get_current_document(browser))
    {
     if(browser->sUri)
     {
      gtk_entry_set_text(GTK_ENTRY(browser->navibar_location), browser->sUri);
      gtk_label_set_text(GTK_LABEL(browser->document_name), browser->sUri);
     }
     update_status_message(browser, NULL);
     update_gui_state(browser);
    }
}

void on_document_title_changed(GtkWidget* document
#ifdef MIDORI_WEBKIT
 , const gchar* sTitle, const gchar* sUri
#endif
 , CBrowser* browser)
{
    #ifdef MIDORI_WEBKIT
    // TODO: We emulate location_changed here for now
    // Shouldn't we have separated title_changed and location_changed signals?
    on_document_location_changed(document, sUri, browser);
    #endif
    DEBUG_FUNC
    #ifndef MIDORI_WEBKIT
    const gchar* sTitle = document_get_title(document);
    #endif
    browser->sTitle = g_strdup(sTitle);
    if(browser->sTitle)
    {
     gtk_label_set_text(GTK_LABEL(browser->document_name), browser->sTitle);
     sokoke_widget_set_tooltip_text(gtk_widget_get_parent(
      gtk_widget_get_parent(browser->document_name)), browser->sTitle);
     gtk_label_set_text(GTK_LABEL(gtk_bin_get_child(GTK_BIN(
      browser->document_menu))), browser->sTitle);
     if(document == get_current_document(browser))
     {
      gchar* sWindowTitle = g_strconcat(browser->sTitle, " - ", PACKAGE_NAME, NULL);
      gtk_window_set_title(GTK_WINDOW(browser->window), sWindowTitle);
      g_free(sWindowTitle);
     }
    }
}

#ifdef MIDORI_WEBKIT
void on_frame_title_changed(WebKitGtkFrame* frame
 , const gchar* sTitle, const gchar* sUri, CBrowser* browser)
{
    DEBUG_FUNC
    UNIMPLEMENTED
}

void on_document_icon_changed(GtkWidget* document, WebKitGtkFrame* widget
 , CBrowser* browser)
{
    DEBUG_FUNC
    // TODO: Implement icon updates; currently this isn't ever called anyway
    const gchar* sIcon = NULL;
    UNIMPLEMENTED
    if(sIcon)
    {
     gtk_label_set_text(GTK_LABEL(browser->document_name), "icon");
     gtk_label_set_text(GTK_LABEL(gtk_bin_get_child(GTK_BIN(
      browser->document_menu))), "icon");
     if(document == get_current_document(browser))
     {
      gchar* sWindowTitle = g_strconcat("icon", " - ", PACKAGE_NAME, NULL);
      gtk_window_set_title(GTK_WINDOW(browser->window), sWindowTitle);
      g_free(sWindowTitle);
     }
    }
}
#endif

void on_document_load_started(GtkWidget* document
#ifdef MIDORI_WEBKIT
 , WebKitGtkFrame* widget
#endif
, CBrowser* browser)
{
    DEBUG_FUNC
    browser->iLoadedPercent = 0;
    browser->iLoadedBytes = 0;
    browser->iLoadedBytesMax = 0;
    update_favicon(browser);
    if(document == get_current_document(browser)) update_gui_state(browser);
    update_statusbar_text(browser);
    gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(browser->progress), 0.1);
    gtk_widget_show(browser->progress);
}

void on_document_load_changed(GtkWidget* document
#ifdef MIDORI_WEBKIT
 , gint progress
#else
 , const WebiLoadStatus* status
#endif
 , CBrowser* browser)
{
    DEBUG_FUNC
    #ifndef MIDORI_WEBKIT
    browser->iLoadedBytes = status->received;
   browser->iLoadedBytesMax = status->size;
    #else
    browser->iLoadedBytes = progress;
    browser->iLoadedBytes = 100;
    #endif
    if(document == get_current_document(browser)) update_gui_state(browser);
}

void on_document_load_finished(GtkWidget* document
#ifdef MIDORI_WEBKIT
 , WebKitGtkFrame* widget
#endif
, CBrowser* browser)
{
    DEBUG_FUNC
    browser->iLoadedPercent = -1;
    update_favicon(browser);
    if(document == get_current_document(browser)) update_gui_state(browser);
    update_statusbar_text(browser);
    gtk_widget_hide(browser->progress);
}

#ifdef MIDORI_WEBKIT
void on_frame_cleared(WebKitGtkFrame* frame, CBrowser* browser)
{
    DEBUG_FUNC
    UNIMPLEMENTED
}

void on_frame_load_done(WebKitGtkFrame* frame, gboolean bSuccess
 , CBrowser* browser)
{
    DEBUG_FUNC
    UNIMPLEMENTED
}
#endif

void on_document_status_message(GtkWidget* document, gchar* sText
 , CBrowser* browser)
{
    DEBUG_FUNC
    update_status_message(browser, sText);
}

void on_document_selection_changed(GtkWidget* document, CBrowser* browser)
{
    //g_print("%s\n", "selection_changed");
}

void on_document_request_script_prompt(GtkWidget* document
#ifdef MIDORI_WEBI
 , WebiPromptArgs* args
#endif
 , CBrowser* browser)
{
    DEBUG_FUNC
    GtkDialog* dialog;
    gchar* title;
    GtkMessageType type = GTK_MESSAGE_OTHER;
    GtkButtonsType buttons = GTK_BUTTONS_NONE;
    gint default_response = GTK_RESPONSE_NONE;
    gint response;
    GtkWidget* widget = NULL;

    #ifdef MIDORI_WEBI
    switch(args->type)
    {
     case WEBI_ALERT:
      type = GTK_MESSAGE_WARNING;
      buttons = GTK_BUTTONS_CLOSE;
      default_response = GTK_RESPONSE_CLOSE; break;
     case WEBI_CONFIRM:
      type = GTK_MESSAGE_QUESTION;
      buttons = GTK_BUTTONS_YES_NO;
      default_response = GTK_RESPONSE_YES; break;
     case WEBI_INPUT:
      type = GTK_MESSAGE_QUESTION;
      buttons = GTK_BUTTONS_OK_CANCEL;
      default_response = GTK_RESPONSE_OK;
    }
    #endif

    dialog = GTK_DIALOG(gtk_message_dialog_new(GTK_WINDOW(browser->window)
     , GTK_DIALOG_DESTROY_WITH_PARENT, type, buttons
    #ifdef MIDORI_WEBI
     , args->msg
    #else
     , ""
    #endif
    ));
    title = g_strconcat("JavaScript - ", browser->sUri, NULL);
    gtk_window_set_title(GTK_WINDOW(dialog), title); g_free(title);

    #ifdef MIDORI_WEBI
    if(args->type == WEBI_INPUT)
    #else
    if(0)
    #endif
    {
     widget = gtk_entry_new();
     #ifdef MIDORI_WEBI
     gtk_entry_set_text(GTK_ENTRY(widget), args->default_input);
     #endif
     gtk_container_add(GTK_CONTAINER(dialog->vbox), widget);
     gtk_widget_show(widget);
    }

    gtk_dialog_set_default_response(dialog, default_response);
    gtk_widget_set_sensitive(document, FALSE);
    response = sokoke_dialog_run_modeless(dialog);
    gtk_widget_set_sensitive(document, TRUE);

    #ifdef MIDORI_WEBI
    switch(response)
    {
     case GTK_RESPONSE_YES:
      args->out_ok_pressed = TRUE; break;
     case GTK_RESPONSE_NO:
      args->out_ok_pressed = FALSE; break;
     case GTK_RESPONSE_OK:
      args->out_input = g_strdup(gtk_entry_get_text(GTK_ENTRY(widget)));
    }
    #endif
    gtk_widget_destroy(GTK_WIDGET(dialog));
}

void on_document_request_auth_prompt(GtkWidget* document
#ifdef MIDORI_WEBI
 , WebiAuthArgs* args
#endif
 , CBrowser* browser)
{
    DEBUG_FUNC
    GtkDialog* dialog;
    gchar* title; gchar* msg;
    GtkWidget* hbox;
    GtkSizeGroup* sizegroup; GtkWidget* align; GtkWidget* label;
    GtkWidget* user; GtkWidget* pass;
    gint response;

    #ifdef MIDORI_WEBI
    msg = g_strdup_printf("Authentication required for:\n\n%s", args->realm);
    #endif
    dialog = GTK_DIALOG(gtk_message_dialog_new(GTK_WINDOW(browser->window)
     , GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_OTHER, GTK_BUTTONS_OK_CANCEL
     , msg)); g_free(msg);
    #ifdef MIDORI_WEBI
    title = g_strconcat("Authentication - ", args->url, NULL);
    #endif
    gtk_window_set_title(GTK_WINDOW(dialog), title); g_free(title);
    #if GTK_CHECK_VERSION(2, 10, 0)
    GtkWidget* image = gtk_image_new_from_icon_name(
     GTK_STOCK_DIALOG_AUTHENTICATION, GTK_ICON_SIZE_DIALOG);
    gtk_widget_show(image);
    gtk_message_dialog_set_image(GTK_MESSAGE_DIALOG(dialog), image);
    #endif
    sizegroup = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);
    hbox = gtk_hbox_new(FALSE, 6);
    label = gtk_label_new("Username");
    align = gtk_alignment_new(0, 0.5, 0, 0);
    gtk_container_add(GTK_CONTAINER(align), label);
    gtk_size_group_add_widget(sizegroup, align);
    gtk_box_pack_start(GTK_BOX(hbox), align, FALSE, FALSE, 0);
    user = gtk_entry_new();
    gtk_box_pack_start(GTK_BOX(hbox), user, FALSE, FALSE, 0);
    gtk_container_add(GTK_CONTAINER(dialog->vbox), hbox);
    gtk_widget_show_all(hbox);
    hbox = gtk_hbox_new(FALSE, 6);
    label = gtk_label_new("Password");
    align = gtk_alignment_new(0, 0.5, 0, 0);
    gtk_container_add(GTK_CONTAINER(align), label);
    gtk_size_group_add_widget(sizegroup, align);
    gtk_box_pack_start(GTK_BOX(hbox), align, FALSE, FALSE, 0);
    pass = gtk_entry_new();
    gtk_entry_set_visibility(GTK_ENTRY(pass), FALSE);
    gtk_box_pack_start(GTK_BOX(hbox), pass, FALSE, FALSE, 0);
    gtk_container_add(GTK_CONTAINER(dialog->vbox), hbox);
    gtk_widget_show_all(hbox);
    gtk_dialog_set_default_response(dialog, GTK_RESPONSE_CANCEL);

    gtk_widget_set_sensitive(document, FALSE);
    response = sokoke_dialog_run_modeless(dialog);
    gtk_widget_set_sensitive(document, TRUE);

    #ifdef MIDORI_WEBI
    if(response == GTK_RESPONSE_OK)
    {
     args->out_username = g_strdup(gtk_entry_get_text(GTK_ENTRY(user)));
     args->out_password = g_strdup(gtk_entry_get_text(GTK_ENTRY(pass)));
     args->out_ok_pressed = TRUE;
    }
    else
     args->out_ok_pressed = FALSE;
    #endif
    gtk_widget_destroy(GTK_WIDGET(dialog));
}

gboolean on_document_set_cookie(GtkWidget* document
#ifdef MIDORI_WEBI
 , WebiCookie* cookie
#endif
 , CBrowser* browser)
{
    DEBUG_FUNC
    // Whole cookie
    // const gchar* cookie;

    // Cookie name
    // const gchar* name;

    // Cookie value
    // const gchar* value;

    // Cookie comment
    // const gchar* comment;

    // Cookie domain
    // const gchar* domain;

    // Cookie TTL
    // const gchar* maxAge;

    // Cookie path
    // const gchar* path;

    // Is cookie secure?
    // gboolean secure;

    // Cookie-specification version.
    // gint version;

    return TRUE;
}

void on_document_link_hover(GtkWidget* document, const gchar* sTitle
 , const gchar* sLabel, const gchar* sUri, const gchar* sTarget, CBrowser* browser)
{
    DEBUG_FUNC
    update_status_message(browser, sUri);
    sokoke_widget_set_visible(browser->icon_link_window
     , sTarget && !g_ascii_strncasecmp(sTarget, "_blank", 6));
    // TODO: detect "onclick", "onmousedown", "onmouseup" javascript links
    sokoke_widget_set_visible(browser->icon_link_script
     , sUri && g_str_has_prefix(sUri, "javascript:"));
}

GtkWidget* on_document_window_open(GtkWidget* document, const gchar* sUri
 , CBrowser* browser)
{
    DEBUG_FUNC
    // A window is created
    // TODO: Find out if this comes from a script or a click
    // TODO: Block scripted popups, return NULL and show status icon
    // Shoud we open a new window or a new tab?
    gboolean bWindow;
    switch(config->iPopupRule)
    {
     case 0: // All popups
         bWindow = TRUE; break;
     case 1: // Only customized popups
         // TODO: Check if window has non-default features
         bWindow = TRUE; break;
     default: // case 2: // No new windows
         bWindow = FALSE;
    }
    CBrowser* newBrowser = browser_new(bWindow ? NULL : browser);
    /*webi_set_group(WEBI(newBrowser->document)
     , webi_get_group(get_current_document(browser)));*/
    return newBrowser->document;
}

void on_document_window_features(GtkWidget* document
#ifdef MIDORI_WEBI
 , WebiWindowProperties* properties
#endif
 , CBrowser* browser)
{
    DEBUG_FUNC
    #ifdef MIDORI_WEBI
    browser->bMenubar = TRUE;
    browser->bToolbar = properties->toolbarsVisible;
    browser->bLocation = properties->toolbarsVisible;
    browser->bStatusbar = properties->statusBarVisible;
    //browser->bScrollbars = properties->scrollbarsVisible;
    browser->bResizable = properties->isResizable;
    #endif
    update_visibility(browser, TRUE);
}

// Document actions begin here

#define DOCUMENT_IS_SENSITIVE(rv) \
 if(!GTK_WIDGET_SENSITIVE(document)) return rv

void document_load_uri(GtkWidget* document, const gchar* sUri)
{
    DEBUG_FUNC
    DOCUMENT_IS_SENSITIVE();
    // TODO: Check if we want an external app for this uri scheme
    // TODO: This isn't the place for uri scheme handling, is it?
    #ifdef MIDORI_WEBI
    webi_load_url(WEBI(document), sUri);
    #elif defined(MIDORI_WEBKIT)
    // TODO: The block below is just for testing :D
    if(sUri && g_str_has_prefix(sUri, "res://"))
    {
     webkit_gtk_page_load_html_string(WEBKIT_GTK_PAGE(document)
      , "<html><head><title>" PACKAGE_STRING "</title></head>"
        "<body>test</body></html>"
      , "about:about");
     return;
    }
    webkit_gtk_page_open(WEBKIT_GTK_PAGE(document), (gchar*)sUri);
    #endif
}

void document_go_back(GtkWidget* document)
{
    DEBUG_FUNC
    DOCUMENT_IS_SENSITIVE();
    #ifdef MIDORI_WEBI
    webi_go_back(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    webkit_gtk_page_go_backward(WEBKIT_GTK_PAGE(document));
    #endif
}

void document_go_forward(GtkWidget* document)
{
    DEBUG_FUNC
    DOCUMENT_IS_SENSITIVE();
    #ifdef MIDORI_WEBI
    webi_go_forward(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    webkit_gtk_page_go_forward(WEBKIT_GTK_PAGE(document));
    #endif
}

void document_refresh(GtkWidget* document, gboolean bSkipCache)
{
    DEBUG_FUNC
    DOCUMENT_IS_SENSITIVE();
    #ifdef MIDORI_WEBI
    webi_refresh(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    webkit_gtk_page_reload(WEBKIT_GTK_PAGE(document));
    #endif
}

void document_stop(GtkWidget* document)
{
    DEBUG_FUNC
    DOCUMENT_IS_SENSITIVE();
    #ifdef MIDORI_WEBI
    webi_stop_load(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    webkit_gtk_page_stop_loading(WEBKIT_GTK_PAGE(document));
    #endif
}

gboolean document_can_go_back(GtkWidget* document)
{
    DEBUG_FUNC
    ZOMBIE_DOC(FALSE)
    #ifdef MIDORI_WEBI
    return webi_can_go_back(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    return webkit_gtk_page_can_go_backward(WEBKIT_GTK_PAGE(document));
    #endif
}

gboolean document_can_go_forward(GtkWidget* document)
{
    DEBUG_FUNC
    ZOMBIE_DOC(FALSE)
    #ifdef MIDORI_WEBI
    return webi_can_go_forward(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    return webkit_gtk_page_can_go_forward(WEBKIT_GTK_PAGE(document));
    #endif
}

G_CONST_RETURN gchar* document_get_uri(GtkWidget* document)
{
    DEBUG_FUNC
    ZOMBIE_DOC(NULL)
    #ifdef MIDORI_WEBI
    return webi_get_location(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    WebKitGtkFrame* frame = webkit_gtk_page_get_main_frame(WEBKIT_GTK_PAGE(document));
    webkit_gtk_frame_get_location(frame);
    #endif
}

G_CONST_RETURN gchar* document_get_title(GtkWidget* document)
{
    DEBUG_FUNC
    ZOMBIE_DOC(NULL)
    #ifdef MIDORI_WEBI
    return webi_get_title(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    WebKitGtkFrame* frame = webkit_gtk_page_get_main_frame(WEBKIT_GTK_PAGE(document));
    webkit_gtk_frame_get_title(frame);
    #endif
}

// FIXME: This better be document_find(document, text, forward, case, wrap)
gboolean document_find(GtkWidget* document, const gchar* sText
 , gboolean bCase, gboolean bForward)
{
    DEBUG_FUNC
    DOCUMENT_IS_SENSITIVE(FALSE);
    #ifdef MIDORI_WEBI
    return webi_find(WEBI(document), sText, bCase, bForward);
    #elif defined(MIDORI_WEBKIT)
    UNIMPLEMENTED return FALSE; /*webkit_gtk_page_search_for(WEBKIT_GTK_PAGE(document)
     , sText, bForward, bCase, TRUE);*/
    #endif
}

gboolean document_can_increase_text_size(GtkWidget* document)
{
    DEBUG_FUNC
    ZOMBIE_DOC(FALSE)
    DOCUMENT_IS_SENSITIVE(FALSE);
    #ifdef MIDORI_WEBI
    return TRUE;
    #elif defined(MIDORI_WEBKIT)
    UNIMPLEMENTED return FALSE; //webkit_gtk_page_can_increase_text_size(WEBKIT_GTK_PAGE(document));
    #endif
}

gboolean document_can_decrease_text_size(GtkWidget* document)
{
    DEBUG_FUNC
    ZOMBIE_DOC(FALSE)
    DOCUMENT_IS_SENSITIVE(FALSE);
    #ifdef MIDORI_WEBI
    return TRUE;
    #elif defined(MIDORI_WEBKIT)
    UNIMPLEMENTED return FALSE; //webkit_gtk_page_can_decrease_text_size(WEBKIT_GTK_PAGE(document));
    #endif
}

void document_set_text_size(GtkWidget* document, gfloat iSize)
{
    DEBUG_FUNC
    DOCUMENT_IS_SENSITIVE();
    #ifdef MIDORI_WEBI
    webi_set_text_multiplier(WEBI(document), iSize);
    #elif defined(MIDORI_WEBKIT)
    UNIMPLEMENTED //webkit_gtk_page_set_text_size_multiplier(WEBKIT_GTK_PAGE(document), iSize);
    #endif
}

gfloat document_get_text_size(GtkWidget* document)
{
    DEBUG_FUNC
    ZOMBIE_DOC(1)
    #ifdef MIDORI_WEBI
    return webi_get_text_multiplier(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    UNIMPLEMENTED return 1;//webkit_gtk_page_get_text_size_multiplier(WEBKIT_GTK_PAGE(document));
    #endif
}

void document_close(GtkWidget* document, CBrowser* browser)
{
    DEBUG_FUNC
    // TODO: Delete old items if > 10 items
    GtkWidget* menuitem = menu_item_new(browser->sTitle
     , GTK_STOCK_FILE, "", NULL, GDK_CONTROL_MASK
     , G_CALLBACK(on_menu_tabsclosed_item_activate), TRUE, browser);
    sokoke_widget_set_tooltip_text(menuitem, browser->sUri);
    g_object_set_data(G_OBJECT(menuitem), "uri", browser->sUri);
    gtk_menu_shell_prepend(GTK_MENU_SHELL(browser->menu_tabsclosed), menuitem);
    gtk_widget_show(menuitem);
    gtk_widget_destroy(browser->document_menu);
    gtk_notebook_remove_page(GTK_NOTEBOOK(browser->documents)
     , get_document_index(document, browser));
    update_browser_actions(browser);
}

#undef DOCUMENT_IS_SENSITIVE
