/* Copyright (C) 2007 Christian Dywan <christian at twotoasts dot de>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. */

#include "midori.h"

// -- GTK+ signal handlers begin here

void on_action_window_new_activate(GtkAction* action, CBrowser* browser)
{
    browser_new(NULL);
}

void on_action_tab_new_activate(GtkAction* action, CBrowser* browser)
{
    browser_new(browser);
    update_browser_actions(browser);
}

void on_action_tab_close_activate(GtkAction* action, CBrowser* browser)
{
    DEBUG_FUNC
    document_close(get_current_document(browser), browser);
}

void on_action_window_close_activate(GtkAction* action, CBrowser* browser)
{
    gtk_widget_destroy(GTK_WIDGET(browser->window));
}

void on_action_quit_activate(GtkAction* action, CBrowser* browser)
{
    gtk_main_quit();
}

void on_action_find_activate(GtkAction* action, CBrowser* browser)
{
    if(GTK_WIDGET_VISIBLE(browser->findbox))
     gtk_widget_hide(browser->findbox);
    else
    {
     gtk_widget_show(browser->findbox);
     gtk_widget_grab_focus(GTK_WIDGET(browser->findbox_text));
    }
}

void on_action_find_next_activate(GtkAction* action, CBrowser* browser)
{
    if(!GTK_WIDGET_VISIBLE(browser->findbox))
     ; // TODO: What if the findbox is hidden?
    const gchar* sText = gtk_entry_get_text(GTK_ENTRY(browser->findbox_text));
    const gboolean bCase = gtk_toggle_tool_button_get_active(
     GTK_TOGGLE_TOOL_BUTTON(browser->findbox_case));
    document_find(get_current_document(browser), sText, bCase, TRUE);
}

void on_action_find_previous_activate(GtkAction* action, CBrowser* browser)
{
    if(!GTK_WIDGET_VISIBLE(browser->findbox))
     ; // TODO: What if the findbox is hidden?
    const gchar* sText = gtk_entry_get_text(GTK_ENTRY(browser->findbox_text));
    const gboolean bCase = gtk_toggle_tool_button_get_active(
     GTK_TOGGLE_TOOL_BUTTON(browser->findbox_case));
    document_find(get_current_document(browser), sText, bCase, FALSE);
}

gboolean on_settings_homepage_focus_out(GtkWidget* widget, GdkEventFocus event
 , CBrowser* browser)
{
    g_free(config->sHomepage);
    config->sHomepage = g_strdup(gtk_entry_get_text(GTK_ENTRY(widget)));
    return FALSE;
}

void on_settings_loadonstartup_changed(GtkWidget* widget, CBrowser* browser)
{
    config->iStartup = gtk_combo_box_get_active(GTK_COMBO_BOX(widget));
}

void on_settings_toolbarstyle_changed(GtkWidget* widget, CBrowser* browser)
{
    config->iToolbarStyle = gtk_combo_box_get_active(GTK_COMBO_BOX(widget));
    gtk_toolbar_set_style(GTK_TOOLBAR(browser->navibar)
     , config_to_toolbarstyle(config->iToolbarStyle));
}

void on_settings_toolbariconsize_toggled(GtkWidget* widget, CBrowser* browser)
{
    config->iToolbarIconSize = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
    gtk_toolbar_set_icon_size(GTK_TOOLBAR(browser->navibar)
     , config_to_toolbariconsize(config->iToolbarIconSize));
}

void on_settings_closebuttonsontabs_toggled(GtkWidget* widget, CBrowser* browser)
{
    config->iTabClose = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
    GList* browsers = browser_list;
    CBrowser* curBrowser;
    do
    {
     curBrowser = (CBrowser*)browsers->data;
     sokoke_widget_set_visible(curBrowser->document_close, config->iTabClose);
    }
    while(browsers = g_list_next(browsers)); g_list_free(browsers);
}

void on_settings_toolbarsearchbox_toggled(GtkWidget* widget, CBrowser* browser)
{
    config->bToolbarSearchbox = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
    sokoke_widget_set_visible(browser->navibar_search, config->bToolbarSearchbox);
}

gboolean on_settings_locationsearch_focus_out(GtkWidget* widget, GdkEventFocus event
 , CBrowser* browser)
{
    g_free(config->sLocationSearch);
    config->sLocationSearch = g_strdup(gtk_entry_get_text(GTK_ENTRY(widget)));
    return FALSE;
}

void on_action_preferences_activate(GtkAction* action, CBrowser* browser)
{
    DEBUG_FUNC
    // Show the preferences dialog. Create it if necessary.
    if(GTK_IS_DIALOG(settings_dialog))
     gtk_window_present(GTK_WINDOW(settings_dialog));
    else
    {
     gchar* sDialogTitle = g_strdup_printf("%s Preferences", PACKAGE_NAME);
     settings_dialog = gtk_dialog_new_with_buttons(sDialogTitle
                                 , GTK_WINDOW(browser->window)
                                 , GTK_DIALOG_DESTROY_WITH_PARENT
                                 , GTK_STOCK_HELP
                                 , GTK_RESPONSE_HELP
                                 , GTK_STOCK_CLOSE
                                 , GTK_RESPONSE_CLOSE
                                 , NULL);
     gtk_window_set_icon_name(GTK_WINDOW(settings_dialog), GTK_STOCK_PREFERENCES);
     gtk_dialog_set_has_separator(GTK_DIALOG(settings_dialog), FALSE);
     // TODO: Implement some kind of help function
     gtk_dialog_set_response_sensitive(GTK_DIALOG(settings_dialog)
      , GTK_RESPONSE_HELP, FALSE); //...
     g_signal_connect(GTK_OBJECT(settings_dialog), "response"
      , GTK_SIGNAL_FUNC(gtk_widget_destroy), settings_dialog); //...
     // TODO: Do we want tooltips for explainations or can we omit that?
     // TODO: We need mnemonics
     GtkWidget* xfce_heading;
     // TODO: "Assignment creates pointer without type conversion"
     if(xfce_heading = sokoke_xfce_header_new(
      gtk_window_get_icon_name(GTK_WINDOW(browser->window)), sDialogTitle))
      gtk_box_pack_start(GTK_BOX(GTK_DIALOG(settings_dialog)->vbox)
      , xfce_heading, FALSE, FALSE, 0);
     g_free(sDialogTitle);
     GtkWidget* notebook = gtk_notebook_new();
     gtk_container_set_border_width(GTK_CONTAINER(notebook), 6);
     GtkSizeGroup* sizegroup = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);
     GtkSizeGroup* sizegroup2 = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);
     GtkWidget* page; GtkWidget* frame; GtkWidget* table; GtkWidget* align;
     GtkWidget* button; GtkWidget* checkbutton; GtkWidget* colorbutton;
     GtkWidget* combobox; GtkWidget* entry; GtkWidget* hbox; GtkWidget* spinbutton;
     #define PAGE_NEW(__label) page = gtk_vbox_new(FALSE, 0);\
      gtk_container_set_border_width(GTK_CONTAINER(page), 5);\
      gtk_notebook_append_page(GTK_NOTEBOOK(notebook), page, gtk_label_new(__label))
     #define FRAME_NEW(__label) frame = sokoke_hig_frame_new(__label);\
      gtk_container_set_border_width(GTK_CONTAINER(frame), 5);\
      gtk_box_pack_start(GTK_BOX(page), frame, FALSE, FALSE, 0);
     #define TABLE_NEW(__rows, __cols) table = gtk_table_new(__rows, __cols, FALSE);\
      gtk_container_set_border_width(GTK_CONTAINER(table), 5);\
      gtk_container_add(GTK_CONTAINER(frame), table);
     #define WIDGET_ADD(__widget, __left, __right, __top, __bottom)\
      gtk_table_attach(GTK_TABLE(table), __widget\
       , __left, __right, __top, __bottom\
       , 0, GTK_FILL, 8, 2)
     #define FILLED_ADD(__widget, __left, __right, __top, __bottom)\
      gtk_table_attach(GTK_TABLE(table), __widget\
       , __left, __right, __top, __bottom\
       , GTK_EXPAND | GTK_FILL, GTK_FILL, 8, 2)
     #define INDENTED_ADD(__widget, __left, __right, __top, __bottom)\
      align = gtk_alignment_new(0, 0.5, 0, 0);\
      gtk_container_add(GTK_CONTAINER(align), __widget);\
      gtk_size_group_add_widget(sizegroup, align);\
      WIDGET_ADD(align, __left, __right, __top, __bottom)
     #define SEMI_INDENTED_ADD(__widget, __left, __right, __top, __bottom)\
      align = gtk_alignment_new(0, 0.5, 0, 0);\
      gtk_container_add(GTK_CONTAINER(align), __widget);\
      gtk_size_group_add_widget(sizegroup2, align);\
      WIDGET_ADD(align, __left, __right, __top, __bottom)
     #define SPANNED_ADD(__widget, __left, __right, __top, __bottom)\
      align = gtk_alignment_new(0, 0.5, 0, 0);\
      gtk_container_add(GTK_CONTAINER(align), __widget);\
      FILLED_ADD(align, __left, __right, __top, __bottom)
     // Page "General"
     PAGE_NEW("General");
     FRAME_NEW("Startup");
     TABLE_NEW(2, 2);
     INDENTED_ADD(gtk_label_new("Load on startup"), 0, 1, 0, 1);
     combobox = gtk_combo_box_new_text();
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , "Blank page", "Homepage"/*, "Last open pages"*/, NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), config->iStartup);
     g_signal_connect(combobox, "changed"
     , G_CALLBACK(on_settings_loadonstartup_changed), browser);
     FILLED_ADD(combobox, 1, 2, 0, 1);
     INDENTED_ADD(gtk_label_new("Homepage"), 0, 1, 1, 2);
     entry = gtk_entry_new();
     gtk_entry_set_text(GTK_ENTRY(entry), config->sHomepage);
     g_signal_connect(entry, "focus-out-event"
     , G_CALLBACK(on_settings_homepage_focus_out), browser);
     FILLED_ADD(entry, 1, 2, 1, 2);
     // TODO: We need something like "use current website"
     FRAME_NEW("Downloads");
     TABLE_NEW(1, 2);
     INDENTED_ADD(gtk_label_new("Download folder"), 0, 1, 0, 1);
     GtkWidget* filebutton = gtk_file_chooser_button_new(
      "Choose download folder", GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER);
     gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(filebutton)
      , g_get_home_dir()); //...
     gtk_widget_set_sensitive(filebutton, FALSE); //...
     FILLED_ADD(filebutton, 1, 2, 0, 1);
     checkbutton = gtk_check_button_new_with_mnemonic
      ("Show a notification window for finished downloads");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     SPANNED_ADD(checkbutton, 0, 2, 1, 2);
     FRAME_NEW("Languages");
     TABLE_NEW(1, 2);
     INDENTED_ADD(gtk_label_new("Preferred languages"), 0, 1, 0, 1);
     entry = gtk_entry_new();
     // TODO: Make sth like get_browser_languages_default filtering encodings and C out
     // TODO: Provide a real ui with real language names (iso-codes)
     const gchar* const* sLanguages = g_get_language_names();
     //for (i = 0; sLanguages[i] != NULL; i++) g_strdup(sLanguages[i]);
     //for (str = strings; *str; str++);
     gchar* sLanguagesPreferred = g_strjoinv(",", (gchar**)sLanguages);
     gtk_entry_set_text(GTK_ENTRY(entry), sLanguagesPreferred/*config->sLanguagesPreferred*/);
     g_free(sLanguagesPreferred);
     gtk_widget_set_sensitive(entry, FALSE); //...
     FILLED_ADD(entry, 1, 2, 0, 1);
     // Page "Appearance"
     PAGE_NEW("Appearance");
     FRAME_NEW("Font settings");
     TABLE_NEW(5, 2);
     INDENTED_ADD(gtk_label_new("Standard font"), 0, 1, 0, 1);
     button = gtk_font_button_new_with_font("Sans 10"/*config->sFontStandard*/);
     gtk_widget_set_sensitive(button, FALSE); //...
     FILLED_ADD(button, 1, 2, 0, 1);
     checkbutton = gtk_check_button_new_with_mnemonic
      ("Use theme _default font");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     SPANNED_ADD(checkbutton, 0, 2, 1, 2);
     INDENTED_ADD(gtk_label_new("Minimum font size"), 0, 1, 2, 3);
     spinbutton = gtk_spin_button_new_with_range(6, 30, 1);
     gtk_spin_button_set_value(GTK_SPIN_BUTTON(spinbutton), 8/*config->iFontSizeMin*/);
     gtk_widget_set_sensitive(spinbutton, FALSE); //...
     SPANNED_ADD(spinbutton, 1, 2, 2, 3);
     INDENTED_ADD(gtk_label_new("Default encoding"), 0, 1, 3, 4);
     combobox = gtk_combo_box_new_text();
     const gchar* encoding = NULL; g_get_charset(&encoding);
     // TODO: Fallback to utf-8 if the encoding is not sane (e.g. when lang=C)
     gchar* sEncodingDefault = g_strdup_printf("Default (%s)", encoding);
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , sEncodingDefault, "Chinese", "Greek", "Japanese (SHIFT_JIS)"
      , "Korean", "Russian", "Unicode (UTF-8)", "Western (ISO-8859-1)", NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), 0); //...
     gtk_widget_set_sensitive(combobox, FALSE); //...
     FILLED_ADD(combobox, 1, 2, 3, 4);
     button = gtk_button_new_with_label("Advanced settings");
     gtk_widget_set_sensitive(button, FALSE); //...
     WIDGET_ADD(button, 1, 2, 3, 4);
     FRAME_NEW("Default colors");
     TABLE_NEW(2, 4);
     SEMI_INDENTED_ADD(gtk_label_new("Text color"), 0, 1, 0, 1);
     colorbutton = gtk_color_button_new();
     gtk_widget_set_sensitive(colorbutton, FALSE); //...
     WIDGET_ADD(colorbutton, 1, 2, 0, 1);
     SEMI_INDENTED_ADD(gtk_label_new("Background color"), 2, 3, 0, 1);
     colorbutton = gtk_color_button_new();
     gtk_widget_set_sensitive(colorbutton, FALSE); //...
     WIDGET_ADD(colorbutton, 3, 4, 0, 1);
     SEMI_INDENTED_ADD(gtk_label_new("Normal link color"), 0, 1, 1, 2);
     colorbutton = gtk_color_button_new();
     gtk_widget_set_sensitive(colorbutton, FALSE); //...
     WIDGET_ADD(colorbutton, 1, 2, 1, 2);
     SEMI_INDENTED_ADD(gtk_label_new("Visited link color"), 2, 3, 1, 2);
     colorbutton = gtk_color_button_new();
     gtk_widget_set_sensitive(colorbutton, FALSE); //...
     WIDGET_ADD(colorbutton, 3, 4, 1, 2);
     // Page "Interface"
     PAGE_NEW("Interface");
     FRAME_NEW("Navigationbar");
     TABLE_NEW(3, 2);
     INDENTED_ADD(gtk_label_new_with_mnemonic("_Toolbar style"), 0, 1, 0, 1);
     combobox = gtk_combo_box_new_text();
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , "Default", "Icons", "Text", "Both", "Both horizontal", NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), config->iToolbarStyle);
     g_signal_connect(combobox, "changed"
     , G_CALLBACK(on_settings_toolbarstyle_changed), browser);
     FILLED_ADD(combobox, 1, 2, 0, 1);
     checkbutton = gtk_check_button_new_with_mnemonic("Show small _icons");
     gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton)
      , config->iToolbarIconSize > 0);
     g_signal_connect(checkbutton, "toggled"
     , G_CALLBACK(on_settings_toolbariconsize_toggled), browser);
     SPANNED_ADD(checkbutton, 0, 1, 1, 2);
     checkbutton = gtk_check_button_new_with_mnemonic("Show _searchbox");
     gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton)
      , config->bToolbarSearchbox);
     g_signal_connect(checkbutton, "toggled"
     , G_CALLBACK(on_settings_toolbarsearchbox_toggled), browser);
     SPANNED_ADD(checkbutton, 1, 2, 1, 2);
     checkbutton = gtk_check_button_new_with_mnemonic("Show _new tab button");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     SPANNED_ADD(checkbutton, 0, 1, 2, 3);
     checkbutton = gtk_check_button_new_with_mnemonic("Show _closed tabs button");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     SPANNED_ADD(checkbutton, 1, 2, 2, 3);
     FRAME_NEW("Miscellaneous");
     TABLE_NEW(3, 2);
     checkbutton = gtk_check_button_new_with_mnemonic
      ("Show close _buttons on tabs");
     gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton)
      , config->iTabClose);
     g_signal_connect(checkbutton, "toggled"
     , G_CALLBACK(on_settings_closebuttonsontabs_toggled), browser);
     SPANNED_ADD(checkbutton, 0, 2, 0, 1);
     INDENTED_ADD(gtk_label_new_with_mnemonic("Tabbar _placement"), 0, 1, 1, 2);
     combobox = gtk_combo_box_new_text();
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , "Left", "Top", "Right", "Bottom", NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), 1); //...
     gtk_widget_set_sensitive(combobox, FALSE); //...
     FILLED_ADD(combobox, 1, 2, 1, 2);
     INDENTED_ADD(gtk_label_new_with_mnemonic("_Location search engine"), 0, 1, 2, 3);
     entry = gtk_entry_new();
     gtk_entry_set_text(GTK_ENTRY(entry), config->sLocationSearch);
     g_signal_connect(entry, "focus-out-event"
     , G_CALLBACK(on_settings_locationsearch_focus_out), browser);
     FILLED_ADD(entry, 1, 2, 2, 3);
     // Page "Network"
     PAGE_NEW("Network");
     FRAME_NEW("Proxy Server");
     gtk_widget_set_sensitive(frame, FALSE); //...
     TABLE_NEW(5, 2);
     checkbutton = gtk_check_button_new_with_mnemonic("_Custom proxy server");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     SPANNED_ADD(checkbutton, 0, 2, 0, 1);
     hbox = gtk_hbox_new(FALSE, 4);
     INDENTED_ADD(gtk_label_new_with_mnemonic("_Host/ Port"), 0, 1, 1, 2);
     gtk_box_pack_start(GTK_BOX(hbox), gtk_entry_new(), FALSE, FALSE, 0);
     gtk_box_pack_start(GTK_BOX(hbox), gtk_spin_button_new_with_range(0, 65535, 1), FALSE, FALSE, 0);
     FILLED_ADD(hbox, 1, 2, 1, 2);
     checkbutton = gtk_check_button_new_with_mnemonic
      ("Proxy requires authentication");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     // TODO The proxy user and pass need to be indented further
     SPANNED_ADD(checkbutton, 0, 2, 2, 3);
     INDENTED_ADD(gtk_label_new("Username"), 0, 1, 3, 4);
     FILLED_ADD(gtk_entry_new(), 1, 2, 3, 4);
     INDENTED_ADD(gtk_label_new("Password"), 0, 1, 4, 5);
     FILLED_ADD(gtk_entry_new(), 1, 2, 4, 5);
     FRAME_NEW("Cache");
     TABLE_NEW(1, 2);
     INDENTED_ADD(gtk_label_new("Cache size"), 0, 1, 0, 1);
     hbox = gtk_hbox_new(FALSE, 4);
     spinbutton = gtk_spin_button_new_with_range(0, 10000, 10);
     gtk_spin_button_set_value(GTK_SPIN_BUTTON(spinbutton), 100/*config->iCacheSize*/);
     gtk_widget_set_sensitive(spinbutton, FALSE); //...
     gtk_box_pack_start(GTK_BOX(hbox), spinbutton, FALSE, FALSE, 0);
     gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new("MB"), FALSE, FALSE, 0);
     button = gtk_button_new_with_label("Clear cache");
     gtk_widget_set_sensitive(button, FALSE); //...
     gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 4);
     SPANNED_ADD(hbox, 1, 2, 0, 1);
     // Page "Privacy"
     PAGE_NEW("Privacy");
     FRAME_NEW("Cookies");
     TABLE_NEW(3, 2);
     INDENTED_ADD(gtk_label_new("Accept cookies"), 0, 1, 0, 1);
     combobox = gtk_combo_box_new_text();
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , "All cookies", "Session cookies", "None", NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), 0); //...
     gtk_widget_set_sensitive(combobox, FALSE); //...
     FILLED_ADD(combobox, 1, 2, 0, 1);
     checkbutton = gtk_check_button_new_with_mnemonic
      ("Allow cookies from the original website only");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     SPANNED_ADD(checkbutton, 0, 2, 1, 2);
     INDENTED_ADD(gtk_label_new("Maximum cookie age"), 0, 1, 2, 3);
     hbox = gtk_hbox_new(FALSE, 4);
     spinbutton = gtk_spin_button_new_with_range(0, 360, 1);
     gtk_spin_button_set_value(GTK_SPIN_BUTTON(spinbutton), 30/*config->iCookieAgeMax*/);
     gtk_widget_set_sensitive(spinbutton, FALSE); //...
     gtk_box_pack_start(GTK_BOX(hbox), spinbutton, FALSE, FALSE, 0);
     gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new("days"), FALSE, FALSE, 0);
     button = gtk_button_new_with_label("View cookies");
     gtk_widget_set_sensitive(button, FALSE); //...
     gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 4);
     SPANNED_ADD(hbox, 1, 2, 2, 3);
     FRAME_NEW("History");
     TABLE_NEW(3, 2);
     checkbutton = gtk_check_button_new_with_mnemonic("Remember my visited pages");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     INDENTED_ADD(checkbutton, 0, 1, 0, 1);
     hbox = gtk_hbox_new(FALSE, 4);
     spinbutton = gtk_spin_button_new_with_range(0, 360, 1);
     gtk_spin_button_set_value(GTK_SPIN_BUTTON(spinbutton), 30/*config->iHistoryAgeMax*/);
     gtk_widget_set_sensitive(spinbutton, FALSE); //...
     gtk_box_pack_start(GTK_BOX(hbox), spinbutton, FALSE, FALSE, 0);
     gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new("days"), FALSE, FALSE, 0);
     SPANNED_ADD(hbox, 1, 2, 0, 1);
     checkbutton = gtk_check_button_new_with_mnemonic
      ("Remember my form inputs");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     SPANNED_ADD(checkbutton, 0, 2, 1, 2);
     checkbutton = gtk_check_button_new_with_mnemonic
      ("Remember my downloaded files");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     SPANNED_ADD(checkbutton, 0, 2, 2, 3);
     // Page "Programs"
     // TODO: We want to see program icons
     // TODO: Do we want to see available programs only?
     // TODO: Is hardcoding programs a good idea?
     // TODO: We should try to get programs from a global place
     // TODO: We need to determine 'default' programs somehow
     // TODO: Should we add an edonkey client?
     PAGE_NEW("Programs");
     FRAME_NEW("External programs");
     TABLE_NEW(5, 2);
     INDENTED_ADD(gtk_label_new("Text editor"), 0, 1, 0, 1);
     combobox = gtk_combo_box_new_text();
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , "Internal", "Default", "geany", "gedit", "medit", "mousepad"
       , "scite", "Other...", NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), 0); //...
     gtk_widget_set_sensitive(combobox, FALSE); //...
     FILLED_ADD(combobox, 1, 2, 0, 1);
     INDENTED_ADD(gtk_label_new("Email client"), 0, 1, 1, 2);
     combobox = gtk_combo_box_new_text();
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , "Default", "Balsa", "Claws Mail", "Evolution", "Thunderbird"
      , "Other...", NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), 0); //...
     gtk_widget_set_sensitive(combobox, FALSE); //...
     FILLED_ADD(combobox, 1, 2, 1, 2);
     INDENTED_ADD(gtk_label_new("Download manager"), 0, 1, 2, 3);
     combobox = gtk_combo_box_new_text();
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , "None", "Gwget", "Urlgfe", "Other...", NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), 0); //...
     gtk_widget_set_sensitive(combobox, FALSE); //...
     FILLED_ADD(combobox, 1, 2, 2, 3);
     INDENTED_ADD(gtk_label_new("BitTorrent client"), 0, 1, 3, 4);
     combobox = gtk_combo_box_new_text();
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , "None", "Azureus", "BitTorrent", "Deluge", "Transmission"
      , "Other...", NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), 0); //...
     gtk_widget_set_sensitive(combobox, FALSE); //...
     FILLED_ADD(combobox, 1, 2, 3, 4);
     INDENTED_ADD(gtk_label_new("Feed reader"), 0, 1, 4, 5);
     combobox = gtk_combo_box_new_text();
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , "None", "Akregator", "Blam!", "RSSOwl", "Straw", "Other...", NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), 0); //...
     gtk_widget_set_sensitive(combobox, FALSE); //...
     FILLED_ADD(combobox, 1, 2, 4, 5);
     gtk_box_pack_start(GTK_BOX(GTK_DIALOG(settings_dialog)->vbox)
      , notebook, FALSE, FALSE, 0);
     gtk_widget_show_all(settings_dialog);
    }
}

void on_action_toolbar_navigation_activate(GtkToggleAction* action, CBrowser* browser)
{
    config->bToolbarNavigation = gtk_toggle_action_get_active(action);
    sokoke_widget_set_visible(browser->navibar, config->bToolbarNavigation);
}

void on_action_toolbar_bookmarks_activate(GtkToggleAction* action, CBrowser* browser)
{
    config->bToolbarBookmarks = gtk_toggle_action_get_active(action);
    sokoke_widget_set_visible(browser->bookmarkbar, config->bToolbarBookmarks);
}

void on_action_toolbar_downloads_activate(GtkToggleAction* action, CBrowser* browser)
{
    /*config->bToolbarDownloads = gtk_toggle_action_get_active(action);
    sokoke_widget_set_visible(browser->downloadbar, config->bToolbarDownloads);*/
}

void on_action_toolbar_status_activate(GtkToggleAction* action, CBrowser* browser)
{
    config->bToolbarStatus = gtk_toggle_action_get_active(action);
    sokoke_widget_set_visible(browser->statusbar, config->bToolbarStatus);
}

void on_action_refresh_activate(GtkAction* action, CBrowser* browser)
{
    GdkModifierType state = (GdkModifierType)0;
    gint x, y; gdk_window_get_pointer(NULL, &x, &y, &state);
    document_refresh(get_current_document(browser), state & GDK_SHIFT_MASK);
}

void on_action_refresh_stop_activate(GtkAction* action, CBrowser* browser)
{
    DEBUG_FUNC
    gchar* sStockId; g_object_get(action, "stock-id", &sStockId, NULL);
    // Refresh or stop, depending on the stock id
    if(!strcmp(sStockId, GTK_STOCK_REFRESH))
    {
     GdkModifierType state = (GdkModifierType)0;
     gint x, y; gdk_window_get_pointer(NULL, &x, &y, &state);
     document_refresh(get_current_document(browser), state & GDK_SHIFT_MASK);
    }
    else
     document_stop(get_current_document(browser));
    g_free(sStockId);
}

void on_action_stop_activate(GtkAction* action, CBrowser* browser)
{
    document_stop(get_current_document(browser));
}

void on_action_zoom_in_activate(GtkAction* action, CBrowser* browser)
{
    const gfloat zoom = document_get_text_size(get_current_document(browser));
    document_set_text_size(get_current_document(browser), zoom + 0.1);
}

void on_action_zoom_out_activate(GtkAction* action, CBrowser* browser)
{
    const gfloat zoom = document_get_text_size(get_current_document(browser));
    document_set_text_size(get_current_document(browser), zoom - 0.1);
}

void on_action_zoom_normal_activate(GtkAction* action, CBrowser* browser)
{
    document_set_text_size(get_current_document(browser), 1);
}

void on_action_back_activate(GtkAction* action, CBrowser* browser)
{
    document_go_back(get_current_document(browser));
}

void on_action_forward_activate(GtkAction* action, CBrowser* browser)
{
    document_go_forward(get_current_document(browser));
}

void on_action_home_activate(GtkAction* action, CBrowser* browser)
{
    document_load_uri(get_current_document(browser), config->sHomepage);
}

void on_action_location_activate(GtkAction* action, CBrowser* browser)
{
    gtk_widget_grab_focus(browser->navibar_location);
    // TODO: What if the navibar is invisible?
}

void on_action_websearch_activate(GtkAction* action, CBrowser* browser)
{
    gtk_widget_grab_focus(browser->navibar_search);
    // TODO: What if the search is invisible?
}

void on_menu_tabsclosed_item_activate(GtkMenuItem* menuitem, CBrowser* browser)
{
    DEBUG_FUNC
    // Create a new document with an uri which has been closed before
    const gchar* sUri = (const gchar*)g_object_get_data(G_OBJECT(menuitem), "uri");
    CBrowser* curBrowser = browser_new(browser);
    document_load_uri(curBrowser->document, sUri);
    gtk_widget_destroy(GTK_WIDGET(menuitem));
    update_browser_actions(curBrowser);
}

void on_action_tabclosed_undo_activate(GtkAction* action, CBrowser* browser)
{
    DEBUG_FUNC
    // Emulate a click on the most recent closedtabs item
    GtkWidget* menuitem;
    GList* children = gtk_container_get_children(
     GTK_CONTAINER(browser->menu_tabsclosed));
    menuitem = (GtkWidget*)g_list_nth_data(children, 0);
    on_menu_tabsclosed_item_activate(GTK_MENU_ITEM(menuitem), browser);
    g_list_free(children);
}

void on_action_tabsclosed_clear_activate(GtkAction* action, CBrowser* browser)
{
    DEBUG_FUNC
    // Clear the closed tabs list, ie. all items that have an uri
    GtkWidget* menuitem; GtkWidget* icon;
    GList* children = gtk_container_get_children(
     GTK_CONTAINER(browser->menu_tabsclosed));
    do
    {
     menuitem = (GtkWidget*)children->data;
     if(g_object_get_data(G_OBJECT(menuitem), "uri"))
      gtk_widget_destroy(menuitem);
    }
    while(children = g_list_next(children)); g_list_free(children);
    update_browser_actions(browser);
}

void on_action_colorpicker_activate(GtkAction* action, CBrowser* browser)
{
    DEBUG_FUNC
    // Show the color picker. Create it if necessary.
    if(GTK_IS_DIALOG(color_picker))
     gtk_window_present(GTK_WINDOW(color_picker));
    else
    {
     color_picker
     = gtk_dialog_new_with_buttons("Color picker"
                                 , GTK_WINDOW(browser->window)
                                 , GTK_DIALOG_DESTROY_WITH_PARENT
                                 , GTK_STOCK_CLOSE
                                 , GTK_RESPONSE_CLOSE
                                 , NULL);
     gtk_window_set_icon_name(GTK_WINDOW(color_picker), GTK_STOCK_COLOR_PICKER);
     gtk_dialog_set_has_separator(GTK_DIALOG(color_picker), TRUE);
     g_signal_connect(GTK_OBJECT(color_picker), "response"
      , GTK_SIGNAL_FUNC(gtk_widget_destroy), color_picker);
     gtk_container_set_border_width(GTK_CONTAINER(GTK_DIALOG(color_picker)->vbox), 5);
     GtkWidget* color_selection = gtk_color_selection_new();
     gtk_container_set_border_width(GTK_CONTAINER(color_selection), 5);
     // TODO: Could we do anything to make the opacity useful?
     gtk_color_selection_set_has_opacity_control(
      GTK_COLOR_SELECTION(color_selection), FALSE);
     // TODO: Is the palette any good?
     gtk_color_selection_set_has_palette(
      GTK_COLOR_SELECTION(color_selection), TRUE);
     gtk_box_pack_start(GTK_BOX(GTK_DIALOG(color_picker)->vbox)
      , color_selection, TRUE, TRUE, 15);
     gtk_box_set_spacing(GTK_BOX(GTK_DIALOG(color_picker)->vbox), 2);
     // TODO: Any useful buttons we could provide?
     gtk_widget_show_all(color_picker);
    }
}

void on_action_tab_previous_activate(GtkAction* action, CBrowser* browser)
{
    gint page = gtk_notebook_get_current_page(GTK_NOTEBOOK(browser->documents));
    gtk_notebook_set_current_page(GTK_NOTEBOOK(browser->documents), page - 1);
}

void on_action_tab_next_activate(GtkAction* action, CBrowser* browser)
{
    // Advance one tab or jump to the first one if we are at the last one
    gint page = gtk_notebook_get_current_page(GTK_NOTEBOOK(browser->documents));
    if(page == gtk_notebook_get_n_pages(GTK_NOTEBOOK(browser->documents)) - 1)
     page = -1;
    gtk_notebook_set_current_page(GTK_NOTEBOOK(browser->documents), page + 1);
}

void on_window_menu_item_activate(GtkImageMenuItem* widget, CBrowser* browser)
{
    gint page = get_document_index(browser->document, browser);
    gtk_notebook_set_current_page(GTK_NOTEBOOK(browser->documents), page);
}

void on_action_about_activate(GtkAction* action, CBrowser* browser)
{
    DEBUG_FUNC
    gtk_show_about_dialog(GTK_WINDOW(browser->window)
     , "logo-icon-name", gtk_window_get_icon_name(GTK_WINDOW(browser->window))
     , "name", PACKAGE_NAME
     , "version", PACKAGE_VERSION
     , "comments", "A lightweight web browser."
     , "copyright", "Copyright © 2007 Christian Dywan"
     , "website", "http://software.twotoasts.de"
     , "authors", credits_authors
     , "documenters", credits_documenters
     , "artists", credits_artists
     , "license", license_gpl_2
     , "wrap-license", TRUE
     //, "translator-credits", _("translator-credits")
     , NULL);
}

gboolean on_navibar_entry_location_key_down(GtkWidget* widget
 , GdkEventKey* event, CBrowser* browser)
{
    DEBUG_FUNC
    const gchar* sUri; CBrowser* curBrowser;
    switch(event->keyval)
    {
     case GDK_Return:
      sUri = gtk_entry_get_text(GTK_ENTRY(widget));
      if(sUri)
      {
       gchar* sNewUri = magic_uri(sUri, config->sLocationSearch);
       // TODO: Use sUrl intermediately when completion is better
       /* TODO Completion should be generated from history, that is
       the uri as well as the title. */
       entry_completion_append(GTK_ENTRY(widget), sUri);
       document_load_uri(get_current_document(browser), sNewUri);
       g_free(sNewUri);
      }
      return TRUE;
     case GDK_Escape:
      if(sUri = document_get_uri(get_current_document(browser)))
       if(strlen(sUri)) gtk_entry_set_text(GTK_ENTRY(widget), sUri);
      return TRUE;
    }
    return FALSE;
}

void on_navibar_entry_search_activate(GtkWidget* widget, CBrowser* browser)
{
    DEBUG_FUNC
    // TODO: use custom search engine, hardcode google for now
    const gchar* sKeywords = gtk_entry_get_text(GTK_ENTRY(widget));
    gchar* sText = g_strconcat("http://www.google.com/search?q=", sKeywords, NULL);
    entry_completion_append(GTK_ENTRY(widget), sKeywords);
    document_load_uri(get_current_document(browser), sText);
    g_free(sText);
}

void on_navibar_tabsclosed_clicked(GtkToolItem* widget, CBrowser* browser)
{
    DEBUG_FUNC
    gtk_menu_popup(GTK_MENU(browser->menu_tabsclosed), NULL, NULL, NULL, NULL, 0
     , gtk_get_current_event_time());
}

void on_action_panels_activate(GtkToggleAction* action, CBrowser* browser)
{
    config->bPanels = gtk_toggle_action_get_active(action);
    sokoke_widget_set_visible(browser->panels, config->bPanels);
}

void on_action_panel_item_activate(GtkRadioAction* action
 , GtkRadioAction* currentAction, CBrowser* browser)
{
    DEBUG_FUNC
    ZOMBIE()
    // TODO: Activating again should hide the contents; how?
    //gint iValue; gint iCurrentValue;
    //g_object_get(G_OBJECT(action), "value", &iValue, NULL);
    //g_object_get(G_OBJECT(currentAction), "value", &iCurrentValue, NULL);
    //GtkWidget* parent = gtk_widget_get_parent(browser->panels_notebook);
    //sokoke_widget_set_visible(parent, iCurrentValue == iValue);
    /*gtk_paned_set_position(GTK_PANED(gtk_widget_get_parent(browser->panels))
     , iCurrentValue == iValue ? config->iPanelPos : 0);*/
    config->iPanelActive = gtk_radio_action_get_current_value(action);
    gint iPage = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(currentAction), "iPage"));
    gtk_notebook_set_current_page(GTK_NOTEBOOK(browser->panels_notebook), iPage);
    gtk_widget_show(browser->panels);
}

// FIXME: Is this the right place for this?
void on_action_openinpanel_activate(GtkAction* action, CBrowser* browser)
{
    DEBUG_FUNC
    GtkWidget* document = get_current_document(browser);
    g_free(config->sPanelPageholder);
    config->sPanelPageholder = g_strdup(document_get_uri(document));
    GtkAction* action_pageholder =
     gtk_action_group_get_action(browser->actiongroup, "PanelPageholder");
    gint value; g_object_get(G_OBJECT(action_pageholder), "value", &value, NULL);
    sokoke_radio_action_set_current_value(GTK_RADIO_ACTION(action_pageholder), value);
    gtk_widget_show(browser->panels);
    document_load_uri(browser->panel_pageholder, config->sPanelPageholder);
}

void on_panels_button_close_clicked(GtkWidget* widget, CBrowser* browser)
{
    config->bPanels = FALSE;
    gtk_widget_hide(browser->panels);
}

gint on_notebook_tab_mouse_down(GtkWidget* widget, GdkEventButton* event
 , CBrowser* browser)
{
    DEBUG_FUNC
    if(event->button == 1 && event->type == GDK_2BUTTON_PRESS)
    {
     // Toggle the label visibility on double click
     GtkWidget* child = gtk_bin_get_child(GTK_BIN(widget));
     GList* children = gtk_container_get_children(GTK_CONTAINER(child));
     child = (GtkWidget*)g_list_nth_data(children, 1);
     gboolean visible = gtk_widget_get_child_visible(GTK_WIDGET(child));
     gtk_widget_set_child_visible(GTK_WIDGET(child), !visible);
     gint a, b; sokoke_widget_get_text_size(browser->document_name, "M", &a, &b);
     gtk_widget_set_size_request(child, !visible
      ? a * config->iTabSize : 0, !visible ? -1 : 0);
     g_list_free(children);
     return TRUE;
    }
    else if(event->button == 2)
    {
     // Close the document on middle click
     document_close(browser->document, browser);
     return TRUE;
    }

    return FALSE;
}

gint on_notebook_tab_close_mouse_down(GtkWidget* widget, GdkEventButton* event
 , CBrowser* browser)
{
    DEBUG_FUNC
    document_close(browser->document, browser);

    return FALSE;
}

void on_notebook_switch_page(GtkWidget* widget, GtkNotebookPage* page
 , guint page_num, CBrowser* browser)
{
    DEBUG_FUNC
    GtkWidget* docbox = gtk_notebook_get_nth_page(
     GTK_NOTEBOOK(browser->documents), page_num);
    GtkWidget* document = gtk_bin_get_child(GTK_BIN(docbox));
    const gchar* sUri = document_get_uri(document);
    if(sUri) gtk_entry_set_text(GTK_ENTRY(browser->navibar_location), sUri);
    const gchar* sTitle = document_get_title(document);
    if(sTitle)
    {
     gchar* sWindowTitle = g_strconcat(sTitle, " - "
      , PACKAGE_NAME, NULL);
     gtk_window_set_title(GTK_WINDOW(browser->window), sWindowTitle);
     g_free(sWindowTitle);
    }
    update_favicon(browser);
    update_security(browser);
    update_gui_state(browser);
    update_statusbar_text(browser);
    update_feeds(browser);
    update_search_engines(browser);
}

// FIXME: This doesn't belong here
void document_popup(GtkWidget* document, GdkEventButton* event, CBrowser* browser)
{
    gboolean bLink = FALSE; // TODO: Did we right-click a link?
    gboolean bImage = FALSE; // TODO: Did we right-click an image?
    gchar* sSelection = NULL; // TODO: The possibly selected text

    action_set_visible("Back", !bLink && !bImage && !sSelection, browser);
    action_set_visible("Forward", !bLink && !bImage && !sSelection, browser);
    action_set_visible("Refresh", !bLink && !bImage && !sSelection, browser);
    action_set_visible("Stop", !bLink && !bImage && !sSelection, browser);
    action_set_visible("RefreshEvery", !bLink && !bImage && !sSelection, browser);

    action_set_visible("LinkTabNew", bLink, browser);
    action_set_visible("LinkTabCurrent", bLink, browser);
    action_set_visible("LinkWindowNew", bLink, browser);

    action_set_visible("LinkBookmarkNew", bLink, browser);
    action_set_visible("LinkSaveAs", bLink, browser);
    action_set_visible("LinkSaveWith", bLink, browser);
    action_set_visible("LinkCopy", bLink, browser);
    action_set_visible("LinkSend", bLink, browser);

    action_set_visible("SelectionLinksNewTabs", sSelection != NULL, browser);
    action_set_visible("SelectionTextTabNew", sSelection != NULL, browser);
    action_set_visible("SelectionTextTabCurrent", sSelection != NULL, browser);
    action_set_visible("SelectionTextWindowNew", sSelection != NULL, browser);

    action_set_visible("ImageViewTabNew", bImage, browser);
    action_set_visible("ImageViewTabCurrent", bImage, browser);
    action_set_visible("ImageSaveAs", bImage, browser);
    action_set_visible("ImageSaveWith", bImage, browser);
    action_set_visible("ImageCopy", bImage, browser);
    action_set_visible("ImageSend", bImage, browser);

    action_set_visible("UndoTabClose", !bLink && !bImage && !sSelection, browser);
    action_set_visible("BookmarkNew", !bLink && !bImage && !sSelection, browser);
    action_set_visible("SaveAs", !bLink && !bImage && !sSelection, browser);
    action_set_visible("Print", !bLink && !bImage && !sSelection, browser);

    action_set_visible("Copy", sSelection != NULL, browser);
    action_set_visible("SelectAll", sSelection != NULL, browser);
    action_set_visible("SelectionSearch", sSelection != NULL, browser);
    action_set_visible("SelectionSearchWith", sSelection != NULL, browser);
    action_set_visible("SelectionSourceView", sSelection != NULL, browser);

    action_set_visible("BackgroundImage", !bLink && !bImage && !sSelection, browser);
    action_set_visible("SourceView", !bLink && !bImage && !sSelection, browser);
    action_set_visible("Properties", !sSelection, browser);

    sokoke_widget_popup(document, GTK_MENU(browser->menu_document), event);
}

gboolean on_document_mouse_down(GtkWidget* widget, GdkEventButton* event
 , CBrowser* browser)
{
    DEBUG_FUNC
    GdkModifierType state = (GdkModifierType)0;
    gint x, y;
    gdk_window_get_pointer(NULL, &x, &y, &state);
    switch(event->button)
    {
     case 1:
      if(state & GDK_SHIFT_MASK)
      {
       // TODO: Open link in new window
       UNIMPLEMENTED return TRUE;
      }
      else if(state & GDK_MOD1_MASK)
      {
       // TODO: Open link in new tab
       UNIMPLEMENTED return TRUE;
      }
      break;
     case 2:
      if(state & GDK_CONTROL_MASK)
      {
       document_set_text_size(get_current_document(browser), 1);
       return TRUE;
      }
      else
      {
       // TODO: Middle click opens new tab
       UNIMPLEMENTED return TRUE;
      }
      break;
     case 3:
      document_popup(get_current_document(browser), event, browser);
      return TRUE;
    }
    return FALSE;
}

void on_document_popup(GtkWidget* document, CBrowser* browser)
{
    DEBUG_FUNC

    document_popup(get_current_document(browser), NULL, browser);
}

gboolean on_document_scroll(GtkWidget* document, GdkEventScroll* event
 , CBrowser* browser)
{
    DEBUG_FUNC
    GdkModifierType state = (GdkModifierType)0;
    gint x, y;
    gdk_window_get_pointer(NULL, &x, &y, &state);
    if(state & GDK_CONTROL_MASK)
    {
     const gfloat size = document_get_text_size(get_current_document(browser));
     if(event->direction == GDK_SCROLL_DOWN)
      document_set_text_size(get_current_document(browser), size + 0.1);
     else if(event->direction == GDK_SCROLL_UP)
      document_set_text_size(get_current_document(browser), size - 0.1);
     return TRUE;
    }
    else
     return FALSE;
}

void on_document_destroy(GtkWidget* widget, CBrowser* browser)
{
    DEBUG_FUNC
    // Update browser list, free memory and possibly quit
    browser_count--;
    GList* tmp_list = g_list_find(browser_list, browser);
    browser_list = g_list_remove_link(browser_list, tmp_list);
    g_free(browser->sUri);
    g_free(browser->sTitle);
    g_free(browser->sStatusMessage);
    if(!browser_count)
    {
     g_object_unref(browser->actiongroup);
     g_object_unref(browser->menu_document);
     gtk_main_quit();
    }
}

void on_findbox_button_close_clicked(GtkWidget* widget, CBrowser* browser)
{
    gtk_widget_hide(browser->findbox);
}

gboolean on_window_destroy(GtkWidget* widget, GdkEvent* event, CBrowser* browser)
{
    DEBUG_FUNC
    gboolean bProceed = TRUE;
    // TODO: What if there are multiple windows?
    // TODO: Smart dialog, à la 'Session?: Save, Discard, Cancel'
    // TODO: Pref startup: session, ask, homepage, blank <-- ask
    // TODO: Pref quit: session, ask, none <-- ask

    if(0 /*browser_count > 1*/)
    {
     GtkDialog* dialog;
     dialog = GTK_DIALOG(gtk_message_dialog_new(GTK_WINDOW(browser->window)
      , GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_WARNING, GTK_BUTTONS_YES_NO
      , "There is more than one tab open. Do you want to close anyway?"));
     gtk_window_set_title(GTK_WINDOW(dialog), PACKAGE_NAME);

     gtk_dialog_set_default_response(dialog, GTK_RESPONSE_YES);
     bProceed = gtk_dialog_run(dialog) == GTK_RESPONSE_YES;

     gtk_widget_destroy(GTK_WIDGET(dialog));
    }
    return !bProceed;
}

// -- Browser creation begins here

// FIXME: This does not belong here
// TODO: Investigate what signature notify::position actually has
gboolean on_panels_notify_position_activate(GtkPaned* paned, CBrowser* browser)
{
    guint pos = gtk_paned_get_position(paned);
    //if(!pos) g_object_set_data(G_OBJECT(paned), "position-null", TRUE);
    config->iPanelPos = pos;
    //g_print("%d %d\n", pos, config->iPanelPos);
    return FALSE;
}

CBrowser* browser_new(CBrowser* oldBrowser)
{
    DEBUG_FUNC
    browser_count++;
    CBrowser* browser = g_new0(CBrowser, 1);
    browser_list = g_list_prepend(browser_list, browser);

    if(!oldBrowser) {

    GtkWidget* label; GtkWidget* hbox;

    // Setup the window
    browser->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    GdkScreen* screen = gtk_window_get_screen(GTK_WINDOW(browser->window));
    gdouble width = gdk_screen_get_width(screen) / 1.7;
    gdouble height = gdk_screen_get_height(screen) / 1.7;
    gtk_window_set_default_size(GTK_WINDOW(browser->window)
     , (gint)width, (gint)height);
    // FIXME: Use custom program icon
    gtk_window_set_icon_name(GTK_WINDOW(browser->window), "web-browser");
    gtk_window_set_title(GTK_WINDOW(browser->window), PACKAGE_NAME);
    gtk_window_add_accel_group(GTK_WINDOW(browser->window), accel_group);
    g_signal_connect(GTK_OBJECT(browser->window), "delete-event"
     , G_CALLBACK(on_window_destroy), browser);
    GtkWidget* vbox = gtk_vbox_new(FALSE, 0);
    gtk_container_add(GTK_CONTAINER(browser->window), vbox);
    gtk_widget_show(vbox);

    // Let us see some ui manager magic
    browser->actiongroup = gtk_action_group_new("Browser");
    gtk_action_group_add_actions(browser->actiongroup, entries, entries_n, browser);
    gtk_action_group_add_toggle_actions(browser->actiongroup
     , toggle_entries, toggle_entries_n, browser);
    gtk_action_group_add_radio_actions(browser->actiongroup
     , refreshevery_entries, refreshevery_entries_n
     , 300, NULL/*G_CALLBACK(activate_refreshevery_period_action)*/, browser);
    gtk_action_group_add_radio_actions(browser->actiongroup
     , panel_entries, panel_entries_n, -1
     , G_CALLBACK(on_action_panel_item_activate), browser);
    GtkUIManager* ui_manager = gtk_ui_manager_new();
    gtk_ui_manager_insert_action_group(ui_manager, browser->actiongroup, 0);
    gtk_window_add_accel_group(GTK_WINDOW(browser->window)
     , gtk_ui_manager_get_accel_group(ui_manager));

    GError* error = NULL;
    if(!gtk_ui_manager_add_ui_from_string(ui_manager, ui_markup, -1, &error))
    {
     // TODO: Should this be a message dialog? When does this happen?
     g_message("User interface couldn't be created: %s", error->message);
     g_error_free(error);
    }

    GtkAction* action;
    // Make all actions except toplevel menus which lack a callback insensitive
    // This will vanish once all actions are implemented
    gint i; for(i = 0; i < entries_n; i++)
    {
     action = gtk_action_group_get_action(browser->actiongroup, entries[i].name);
     gtk_action_set_sensitive(action, entries[i].callback || !entries[i].tooltip);
    }
    for(i = 0; i < toggle_entries_n; i++)
    {
     action = gtk_action_group_get_action(browser->actiongroup
      , toggle_entries[i].name);
     gtk_action_set_sensitive(action, toggle_entries[i].callback != NULL);
    }
    for(i = 0; i < refreshevery_entries_n; i++)
    {
     action = gtk_action_group_get_action(browser->actiongroup
      , refreshevery_entries[i].name); gtk_action_set_sensitive(action, FALSE);
    }

    action_set_sensitive("PrivateBrowsing", FALSE, browser);
    action_set_sensitive("WorkOffline", FALSE, browser);

    //action_set_active("Menubar", config->bMenubar, browser);
    action_set_active("ToolbarNavigation", config->bToolbarNavigation, browser);
    action_set_active("ToolbarBookmarks", config->bToolbarBookmarks, browser);
    //action_set_active("ToolbarDownloads", config->bToolbarDownloads, browser);
    action_set_active("ToolbarStatus", config->bToolbarStatus, browser);
    action_set_active("Panels", config->bPanels, browser);
    action = gtk_action_group_get_action(browser->actiongroup, "PanelDownloads");
    sokoke_radio_action_set_current_value(GTK_RADIO_ACTION(action), config->iPanelActive);

    // Create the menubar
    browser->menubar = gtk_ui_manager_get_widget(ui_manager, "/menubar");
    GtkWidget* menuitem = gtk_menu_item_new();
    gtk_widget_show(menuitem);
    browser->throbber = gtk_image_new_from_stock(GTK_STOCK_EXECUTE, GTK_ICON_SIZE_MENU);
    //gtk_widget_show(browser->throbber);//...
    gtk_container_add(GTK_CONTAINER(menuitem), browser->throbber);
    gtk_widget_set_sensitive(menuitem, FALSE);
    gtk_menu_item_set_right_justified(GTK_MENU_ITEM(menuitem), TRUE);
    gtk_menu_shell_append(GTK_MENU_SHELL(browser->menubar), menuitem);
    gtk_box_pack_start(GTK_BOX(vbox), browser->menubar, FALSE, FALSE, 0);
    browser->menu_tabsclosed = gtk_ui_manager_get_widget(ui_manager, "/popup_tabsclosed");
    menuitem = gtk_ui_manager_get_widget(ui_manager, "/menubar/Go/TabsClosed");
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuitem), browser->menu_tabsclosed);
    menuitem = gtk_separator_menu_item_new();
    gtk_widget_show(menuitem);
    gtk_menu_shell_prepend(GTK_MENU_SHELL(browser->menu_tabsclosed), menuitem);
    browser->menu_window = gtk_menu_item_get_submenu(
     GTK_MENU_ITEM(gtk_ui_manager_get_widget(ui_manager, "/menubar/Window")));
    menuitem = gtk_separator_menu_item_new();
    gtk_widget_show(menuitem);
    gtk_menu_shell_append(GTK_MENU_SHELL(browser->menu_window), menuitem);
    browser->menu_document = gtk_ui_manager_get_widget(ui_manager, "/popup_document");
    g_object_ref(browser->menu_document);

    // Create the navigation toolbar
    browser->navibar = gtk_ui_manager_get_widget(ui_manager, "/toolbar_navigation");
    gtk_toolbar_set_style(GTK_TOOLBAR(browser->navibar)
     , config_to_toolbarstyle(config->iToolbarStyle));
    gtk_toolbar_set_icon_size(GTK_TOOLBAR(browser->navibar)
     , config_to_toolbariconsize(config->iToolbarIconSize));
    gtk_toolbar_set_show_arrow(GTK_TOOLBAR(browser->navibar), TRUE);
    gtk_box_pack_start(GTK_BOX(vbox), browser->navibar, FALSE, FALSE, 0);
    action = gtk_action_group_get_action(browser->actiongroup, "Back");
    g_object_set(action, "is-important", TRUE, NULL);

    // Location entry
    browser->navibar_location = sexy_icon_entry_new();
    entry_setup_completion(GTK_ENTRY(browser->navibar_location));
    sokoke_entry_set_can_undo(GTK_ENTRY(browser->navibar_location), TRUE);
    browser->icon_page = gtk_image_new();
    sexy_icon_entry_set_icon(SEXY_ICON_ENTRY(browser->navibar_location)
     , SEXY_ICON_ENTRY_PRIMARY, GTK_IMAGE(browser->icon_page));
    sexy_icon_entry_add_clear_button(SEXY_ICON_ENTRY(browser->navibar_location));
    g_signal_connect(browser->navibar_location, "key-press-event"
     , G_CALLBACK(on_navibar_entry_location_key_down), browser);
    GtkToolItem* toolitem = gtk_tool_item_new();
    gtk_tool_item_set_expand(GTK_TOOL_ITEM(toolitem), TRUE);
    gtk_container_add(GTK_CONTAINER(toolitem), browser->navibar_location);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->navibar), toolitem, -1);

    // Search entry
    // TODO: Paint an arrow over the icon
    // Make this actively resizable or enlarge to fit contents?
    // TODO: custom icon
    // select search engine via popup or keyboard or mouse wheel
    // checkbutton 'Open in new tab' in the popup
    browser->navibar_search = sexy_icon_entry_new();
    // TODO: Display "show in context menu" search engines as "completion actions"
    entry_setup_completion(GTK_ENTRY(browser->navibar_search));
    sokoke_entry_set_can_undo(GTK_ENTRY(browser->navibar_search), TRUE);
    sokoke_entry_set_default_text(GTK_ENTRY(browser->navibar_search)
     , "Search the web");
    browser->icon_search = gtk_image_new_from_stock(GTK_STOCK_FIND
     , GTK_ICON_SIZE_MENU);
    sexy_icon_entry_set_icon(SEXY_ICON_ENTRY(browser->navibar_search)
     , SEXY_ICON_ENTRY_PRIMARY, GTK_IMAGE(browser->icon_search));
    browser->icon_search_add = gtk_image_new();
    sexy_icon_entry_set_icon(SEXY_ICON_ENTRY(browser->navibar_search)
     , SEXY_ICON_ENTRY_SECONDARY, GTK_IMAGE(browser->icon_search_add));
    g_signal_connect(browser->navibar_search, "activate"
     , G_CALLBACK(on_navibar_entry_search_activate), browser);
    toolitem = gtk_tool_item_new();
    gtk_container_add(GTK_CONTAINER(toolitem), browser->navibar_search);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->navibar), toolitem, -1);
    action = gtk_action_group_get_action(browser->actiongroup, "TabsClosed");
    toolitem = GTK_TOOL_ITEM(gtk_action_create_tool_item(action));
    g_signal_connect(toolitem, "clicked", G_CALLBACK(on_navibar_tabsclosed_clicked), browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->navibar), toolitem, -1);
    sokoke_container_show_children(GTK_CONTAINER(browser->navibar));

    // Bookmarkbar
    browser->bookmarkbar = gtk_toolbar_new();
    gtk_toolbar_set_icon_size(GTK_TOOLBAR(browser->bookmarkbar), GTK_ICON_SIZE_MENU);
    gtk_toolbar_set_style(GTK_TOOLBAR(browser->bookmarkbar), GTK_TOOLBAR_BOTH_HORIZ);
    toolitem = tool_button_new("Folder", GTK_STOCK_DIRECTORY, TRUE, TRUE
     , NULL, NULL, browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->bookmarkbar), toolitem, -1);
    toolitem = tool_button_new("Example", STOCK_BOOKMARK, TRUE, TRUE
     , NULL/*on_menu_bookmarks_item_activate*/, "http://example.com", browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->bookmarkbar), toolitem, -1);
    toolitem = tool_button_new("Example", STOCK_BOOKMARK, TRUE, TRUE
     , NULL/*on_menu_bookmarks_item_activate*/, "http://example.com", browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->bookmarkbar), toolitem, -1);
    sokoke_container_show_children(GTK_CONTAINER(browser->bookmarkbar));
    gtk_box_pack_start(GTK_BOX(vbox), browser->bookmarkbar, FALSE, FALSE, 0);

    // Superuser warning
    // TODO: "Assignment creates pointer without type conversion"
    if(hbox = sokoke_superuser_warning_new())
     gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

    // Create the panels
    GtkWidget* hpaned = gtk_hpaned_new();
    gtk_paned_set_position(GTK_PANED(hpaned), config->iPanelPos);
    g_signal_connect(hpaned, "notify::position"
     , GTK_SIGNAL_FUNC(on_panels_notify_position_activate), browser);
    gtk_box_pack_start(GTK_BOX(vbox), hpaned, TRUE, TRUE, 0);
    gtk_widget_show(hpaned);

    browser->panels = gtk_hbox_new(FALSE, 0);
    gtk_paned_pack1(GTK_PANED(hpaned), browser->panels, FALSE, FALSE);
    sokoke_widget_set_visible(browser->panels, config->bPanels);

    // Create the panel toolbar
    GtkWidget* panelbar = gtk_ui_manager_get_widget(ui_manager, "/toolbar_panels");
    gtk_toolbar_set_style(GTK_TOOLBAR(panelbar), GTK_TOOLBAR_BOTH);
    gtk_toolbar_set_icon_size(GTK_TOOLBAR(panelbar), GTK_ICON_SIZE_BUTTON);
    gtk_toolbar_set_orientation(GTK_TOOLBAR(panelbar), GTK_ORIENTATION_VERTICAL); 
    gtk_box_pack_start(GTK_BOX(browser->panels), panelbar, FALSE, FALSE, 0);
    //gtk_widget_show(gtk_widget_get_parent(panelbar));
    g_object_unref(ui_manager);

    GtkWidget* cbox = gtk_vbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(browser->panels), cbox, TRUE, TRUE, 0);
    gtk_widget_show(cbox);

    // Panels title bar
    GtkWidget* labelbar = gtk_toolbar_new();
    gtk_toolbar_set_icon_size(GTK_TOOLBAR(labelbar), GTK_ICON_SIZE_MENU);
    gtk_toolbar_set_style(GTK_TOOLBAR(labelbar), GTK_TOOLBAR_ICONS);
    toolitem = gtk_tool_item_new();
    gtk_tool_item_set_expand(toolitem, TRUE);
    label = gtk_label_new_with_mnemonic("_Panels");
    gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
    gtk_container_add(GTK_CONTAINER(toolitem), label);
    gtk_container_set_border_width(GTK_CONTAINER(toolitem), 6);
    gtk_toolbar_insert(GTK_TOOLBAR(labelbar), toolitem, -1);
    // TODO: Does 'goto top' actually indicate 'detach'?
    toolitem = tool_button_new(NULL, GTK_STOCK_GOTO_TOP, FALSE, TRUE
     , NULL/*G_CALLBACK(on_panels_button_float_clicked)*/, "Detach panel", browser);
    gtk_toolbar_insert(GTK_TOOLBAR(labelbar), toolitem, -1);
    toolitem = tool_button_new(NULL, GTK_STOCK_CLOSE, FALSE, TRUE
     , G_CALLBACK(on_panels_button_close_clicked), "Close panel", browser);
    gtk_toolbar_insert(GTK_TOOLBAR(labelbar), toolitem, -1);
    gtk_box_pack_start(GTK_BOX(cbox), labelbar, FALSE, FALSE, 0);
    gtk_widget_show_all(labelbar);

    // Notebook, containing all panels
    browser->panels_notebook = gtk_notebook_new();
    gtk_notebook_set_show_border(GTK_NOTEBOOK(browser->panels_notebook), FALSE);
    gtk_notebook_set_show_tabs(GTK_NOTEBOOK(browser->panels_notebook), FALSE);
      GtkWidget* scrolled; GtkWidget* panel; gint iPage;
      // Dummy: This is the "fallback" panel for now
      iPage = gtk_notebook_append_page(GTK_NOTEBOOK(
       browser->panels_notebook), gtk_label_new("empty"), NULL);
      // Pageholder
      #ifdef MIDORI_WEBKIT
      scrolled = gtk_scrolled_window_new(NULL, NULL);
      gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled)
       , GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
      GTK_WIDGET_SET_FLAGS(scrolled, GTK_CAN_FOCUS);
      panel = webkit_gtk_page_new();
      #else
      scrolled = gtk_event_box_new();
      panel = webi_new();
      #endif
      browser->panel_pageholder = panel;
      //sokoke_widget_set_transparency(browser->window, 0.7);
      gtk_container_add(GTK_CONTAINER(scrolled), panel);
      document_load_uri(panel, config->sPanelPageholder);
      iPage = gtk_notebook_append_page(GTK_NOTEBOOK(
       browser->panels_notebook), scrolled, NULL);
      action = gtk_action_group_get_action(browser->actiongroup, "PanelPageholder");
      g_object_set_data(G_OBJECT(action), "iPage", GINT_TO_POINTER(iPage));
    gtk_box_pack_start(GTK_BOX(cbox), sokoke_frame_new_child_shadow(
     browser->panels_notebook, GTK_SHADOW_IN), TRUE, TRUE, 0);
    //gtk_notebook_set_current_page(GTK_NOTEBOOK(browser->panels_notebook), config->iPanelActive);
    gtk_widget_show_all(gtk_widget_get_parent(browser->panels_notebook));

    // Notebook, containing all documents
    browser->documents = gtk_notebook_new();
    gtk_notebook_set_scrollable(GTK_NOTEBOOK(browser->documents), TRUE);
    #if GTK_CHECK_VERSION(2, 10, 0)
    //gtk_notebook_set_group_id(GTK_NOTEBOOK(browser->documents), 0);
    #endif
    g_signal_connect_after(GTK_OBJECT(browser->documents), "switch-page"
     , GTK_SIGNAL_FUNC(on_notebook_switch_page), browser);
    gtk_paned_pack2(GTK_PANED(hpaned), browser->documents, FALSE, FALSE);
    gtk_widget_show(browser->documents);

    // Incremental findbar
    browser->findbox = gtk_toolbar_new();
    gtk_toolbar_set_icon_size(GTK_TOOLBAR(browser->findbox), GTK_ICON_SIZE_MENU);
    gtk_toolbar_set_style(GTK_TOOLBAR(browser->findbox), GTK_TOOLBAR_BOTH_HORIZ);
    toolitem = gtk_tool_item_new();
    gtk_container_set_border_width(GTK_CONTAINER(toolitem), 6);
    gtk_container_add(GTK_CONTAINER(toolitem)
     , gtk_label_new_with_mnemonic("_Inline find:"));
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), toolitem, -1);
    browser->findbox_text = sexy_icon_entry_new();
    // TODO: Change that to an error icon if a search fails
    GtkWidget* icon = gtk_image_new_from_stock(GTK_STOCK_FIND, GTK_ICON_SIZE_MENU);
    sexy_icon_entry_set_icon(SEXY_ICON_ENTRY(browser->findbox_text)
     , SEXY_ICON_ENTRY_PRIMARY, GTK_IMAGE(icon));
    sexy_icon_entry_add_clear_button(SEXY_ICON_ENTRY(browser->findbox_text));
    sokoke_entry_set_can_undo(GTK_ENTRY(browser->findbox_text), TRUE);
    g_signal_connect(browser->findbox_text, "activate"
     , GTK_SIGNAL_FUNC(on_action_find_next_activate), browser);
    toolitem = gtk_tool_item_new();
    gtk_container_add(GTK_CONTAINER(toolitem), browser->findbox_text);
    gtk_tool_item_set_expand(GTK_TOOL_ITEM(toolitem), TRUE);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), toolitem, -1);
    toolitem = tool_button_new(NULL, GTK_STOCK_GO_BACK, TRUE, TRUE
     , G_CALLBACK(on_action_find_previous_activate), NULL, browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), toolitem, -1);
    toolitem = tool_button_new(NULL, GTK_STOCK_GO_FORWARD, TRUE, TRUE
     , G_CALLBACK(on_action_find_next_activate), NULL, browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), toolitem, -1);
    browser->findbox_case = gtk_toggle_tool_button_new_from_stock(GTK_STOCK_SPELL_CHECK);
    gtk_tool_button_set_label(GTK_TOOL_BUTTON(browser->findbox_case), "Match case");
    gtk_tool_item_set_is_important(GTK_TOOL_ITEM(browser->findbox_case), TRUE);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), browser->findbox_case, -1);
    toolitem = gtk_separator_tool_item_new();
    gtk_separator_tool_item_set_draw(GTK_SEPARATOR_TOOL_ITEM(toolitem), FALSE);
    gtk_tool_item_set_expand(GTK_TOOL_ITEM(toolitem), TRUE);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), toolitem, -1);
    toolitem = tool_button_new(NULL, GTK_STOCK_CLOSE, FALSE, TRUE
     , G_CALLBACK(on_findbox_button_close_clicked), "Close findbar", browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), toolitem, -1);
    sokoke_container_show_children(GTK_CONTAINER(browser->findbox));
    gtk_box_pack_start(GTK_BOX(vbox), browser->findbox, FALSE, FALSE, 0);

    // TODO: fix children overlapping statusbar border
    browser->statusbar = gtk_statusbar_new();
    gtk_box_pack_start(GTK_BOX(vbox), browser->statusbar, FALSE, FALSE, 0);
    browser->progress = gtk_progress_bar_new();
    // Setting the progressbar's height to 1 makes it fit in the statusbar
    gtk_widget_set_size_request(browser->progress, -1, 1);
    gtk_box_pack_start(GTK_BOX(browser->statusbar), browser->progress
     , FALSE, FALSE, 3);

    browser->icon_link_script = gtk_image_new_from_icon_name(
     STOCK_SCRIPT, GTK_ICON_SIZE_MENU);
    gtk_box_pack_start(GTK_BOX(browser->statusbar), browser->icon_link_script
     , FALSE, FALSE, 0);
    browser->icon_link_window = gtk_image_new_from_icon_name(
     STOCK_WINDOW_NEW, GTK_ICON_SIZE_MENU);
    gtk_box_pack_start(GTK_BOX(browser->statusbar), browser->icon_link_window
     , FALSE, FALSE, 0);
    // TODO: Show network status here if possible
    browser->icon_network = gtk_image_new_from_icon_name(
     STOCK_NETWORK_OFFLINE, GTK_ICON_SIZE_MENU);
    gtk_widget_set_sensitive(browser->icon_network, FALSE); //...
    gtk_widget_show(browser->icon_network);
    gtk_box_pack_start(GTK_BOX(browser->statusbar), browser->icon_network
     , FALSE, FALSE, 0);
    browser->icon_security = gtk_image_new();
    gtk_box_pack_start(GTK_BOX(browser->statusbar), browser->icon_security
     , FALSE, FALSE, 0);
    gtk_widget_show(browser->icon_security);
    browser->icon_newsfeed = gtk_image_new_from_icon_name(STOCK_NEWSFEED
     , GTK_ICON_SIZE_MENU);
    gtk_box_pack_start(GTK_BOX(browser->statusbar), browser->icon_newsfeed
     , FALSE, FALSE, 0);

    }
    else
    {

    browser->window = oldBrowser->window;
    browser->menubar = oldBrowser->menubar;
    browser->actiongroup = oldBrowser->actiongroup;
    browser->menu_tabsclosed = oldBrowser->menu_tabsclosed;
    browser->menu_window = oldBrowser->menu_window;
    browser->menu_document = oldBrowser->menu_document;
    browser->throbber = oldBrowser->throbber;
    browser->navibar = oldBrowser->navibar;
    browser->icon_page = oldBrowser->icon_page;
    browser->navibar_location = oldBrowser->navibar_location;
    browser->icon_search = oldBrowser->icon_search;
    browser->icon_search_add = oldBrowser->icon_search_add;
    browser->navibar_search = oldBrowser->navibar_search;
    browser->bookmarkbar = oldBrowser->bookmarkbar;
    browser->panels = oldBrowser->panels;
    browser->panels_notebook = oldBrowser->panels_notebook;
    browser->panel_pageholder = oldBrowser->panel_pageholder;
    browser->documents = oldBrowser->documents;
    browser->findbox = oldBrowser->findbox;
    browser->statusbar = oldBrowser->statusbar;
    browser->icon_link_script = oldBrowser->icon_link_script;
    browser->icon_link_window = oldBrowser->icon_link_window;
    browser->icon_network = oldBrowser->icon_network;
    browser->progress = oldBrowser->progress;
    browser->icon_security = oldBrowser->icon_security;
    browser->icon_newsfeed = oldBrowser->icon_newsfeed;

    }

    // Define some default values
    browser->bMenubar = TRUE;
    browser->bToolbar = TRUE;
    browser->bLocation = TRUE;
    browser->bStatusbar = TRUE;
    browser->bResizable = TRUE;
    browser->sUri = g_strdup("");
    browser->sTitle = g_strdup("about:blank");
    browser->iLoadedPercent = -1; // initially "not loading"

    // Add a window menu item
    // TODO: Menu items should be ordered like the notebook tabs
    // TODO: Watch tab reordering in >= gtk 2.10
    browser->document_menu = menu_item_new(browser->sTitle, GTK_STOCK_FILE, "", NULL
     , (GdkModifierType)0, G_CALLBACK(on_window_menu_item_activate), TRUE, browser);
    gtk_widget_show(browser->document_menu);
    gtk_menu_shell_append(GTK_MENU_SHELL(browser->menu_window), browser->document_menu);

    // Create a new page
    GtkWidget* eventbox = gtk_event_box_new();
    gtk_event_box_set_visible_window(GTK_EVENT_BOX(eventbox), FALSE);
    g_signal_connect(GTK_OBJECT(eventbox), "button_press_event"
     , GTK_SIGNAL_FUNC(on_notebook_tab_mouse_down), browser);
    GtkWidget* hbox = gtk_hbox_new(FALSE, 1);
    gtk_container_add(GTK_CONTAINER(eventbox), GTK_WIDGET(hbox));
    browser->document_icon = gtk_image_new_from_stock(GTK_STOCK_FILE, GTK_ICON_SIZE_MENU);
    gtk_box_pack_start(GTK_BOX(hbox), browser->document_icon, FALSE, FALSE, 0);
    browser->document_name = gtk_label_new(browser->sTitle);
    gtk_misc_set_alignment(GTK_MISC(browser->document_name), 0.0, 0.5);
    // TODO: make the tab initially look "unvisited" until it's focused
    // TODO: tabs should shrink when there is not enough space
    // TODO: gtk's tab scrolling is weird?
    gint w, h; sokoke_widget_get_text_size(browser->document_name, "M", &w, &h);
    gtk_widget_set_size_request(GTK_WIDGET(browser->document_name)
     , w * config->iTabSize, -1);
    gtk_label_set_ellipsize(GTK_LABEL(browser->document_name), PANGO_ELLIPSIZE_END);
    gtk_box_pack_start(GTK_BOX(hbox), browser->document_name, FALSE, FALSE, 0);
    browser->document_close = gtk_button_new();
    gtk_button_set_relief(GTK_BUTTON(browser->document_close), GTK_RELIEF_NONE);
    gtk_button_set_focus_on_click(GTK_BUTTON(browser->document_close), FALSE);
    GtkRcStyle* rcstyle = gtk_rc_style_new();
    rcstyle->xthickness = rcstyle->ythickness = 0;
    gtk_widget_modify_style(browser->document_close, rcstyle);
    gtk_rc_style_unref(rcstyle);
    GtkWidget* image = gtk_image_new_from_stock(GTK_STOCK_CLOSE, GTK_ICON_SIZE_MENU);
    gtk_button_set_image(GTK_BUTTON(browser->document_close), image);
    gtk_box_pack_start(GTK_BOX(hbox), browser->document_close, FALSE, FALSE, 0);
    GtkSettings* gtksettings = gtk_settings_get_default(); gint height;
    gtk_icon_size_lookup_for_settings(gtksettings, GTK_ICON_SIZE_BUTTON, 0, &height);
    gtk_widget_set_size_request(browser->document_close, -1, height);
    gtk_widget_show_all(GTK_WIDGET(eventbox));
    sokoke_widget_set_visible(browser->document_close, config->iTabClose);
    g_signal_connect(GTK_OBJECT(browser->document_close), "button_press_event"
     , GTK_SIGNAL_FUNC(on_notebook_tab_close_mouse_down), browser);
    #ifdef MIDORI_WEBKIT
    GtkWidget* scrolled = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled)
     , GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    GTK_WIDGET_SET_FLAGS(scrolled, GTK_CAN_FOCUS);
    #else
    GtkWidget* scrolled = gtk_event_box_new();
    #endif
    g_signal_connect(GTK_OBJECT(scrolled), "button_press_event"
     , GTK_SIGNAL_FUNC(on_document_mouse_down), browser);
    g_signal_connect(GTK_OBJECT(scrolled), "popup-menu"
     , GTK_SIGNAL_FUNC(on_document_popup), browser);
    g_signal_connect(GTK_OBJECT(scrolled), "scroll-event"
     , GTK_SIGNAL_FUNC(on_document_scroll), browser);
    gint page = gtk_notebook_insert_page(GTK_NOTEBOOK(browser->documents)
     , scrolled, GTK_WIDGET(eventbox)
     , gtk_notebook_get_current_page(GTK_NOTEBOOK(browser->documents)) + 1);
    #if GTK_CHECK_VERSION(2, 10, 0)
    gtk_notebook_set_tab_reorderable(GTK_NOTEBOOK(browser->documents)
     , scrolled, TRUE);
    gtk_notebook_set_tab_detachable(GTK_NOTEBOOK(browser->documents)
     , scrolled, TRUE);
    #endif

    // Create a new document
    #define DOC_CONNECT(__sig, __func) g_signal_connect(\
     G_OBJECT(browser->document), __sig\
     , G_CALLBACK(__func), browser);
    gchar* sAgentName = g_strconcat(PACKAGE_NAME, "/", PACKAGE_VERSION, NULL);
    #ifdef MIDORI_WEBI
    browser->document = webi_new();
    webi_set_emit_internal_status(WEBI(browser->document), TRUE); // crap
    gtk_widget_show_all(browser->document); // bug
    DOC_CONNECT("location", on_document_location_changed)
    DOC_CONNECT("title", on_document_title_changed)
    DOC_CONNECT("load-start", on_document_load_started)
    DOC_CONNECT("load-stop", on_document_load_finished)
    DOC_CONNECT("status", on_document_load_changed)
    DOC_CONNECT("selection", on_document_selection_changed)
    DOC_CONNECT("req-js-prompt", on_document_request_script_prompt)
    DOC_CONNECT("req-auth-prompt", on_document_request_auth_prompt)
    DOC_CONNECT("req-new-window", on_document_window_open)
    DOC_CONNECT("set-window-properties", on_document_window_features)
    DOC_CONNECT("set-cookie", on_document_set_cookie)
    DOC_CONNECT("mouse-over", on_document_link_hover)
    #elif defined(MIDORI_WEBKIT)
    browser->document = GTK_WIDGET(webkit_gtk_page_new());
    WebKitGtkFrame* mainframe = webkit_gtk_page_get_main_frame(
     WEBKIT_GTK_PAGE(browser->document));
    #define FRAME_CONNECT(__sig, __func) g_signal_connect(\
     G_OBJECT(mainframe), __sig\
     , G_CALLBACK(__func), browser);
    /*webkit_gtk_page_set_application_name_for_user_agent(
     PACKAGE_NAME, PACKAGE_VERSION);*/
    //webkit_gtk_page_set_custom_user_agent(sAgentName);
    FRAME_CONNECT("title-changed", on_frame_title_changed)
    DOC_CONNECT("title-changed", on_document_title_changed)
    DOC_CONNECT("icon-loaded", on_document_icon_changed)
    DOC_CONNECT("load-started", on_document_load_started)
    DOC_CONNECT("load-progress-changed", on_document_load_changed)
    DOC_CONNECT("load-finished", on_document_load_finished)
    FRAME_CONNECT("cleared", on_frame_cleared)
    FRAME_CONNECT("load-done", on_frame_load_done)
    DOC_CONNECT("status-bar-text-changed", on_document_status_message)
    #endif
    g_free(sAgentName);

    DOC_CONNECT("destroy", on_document_destroy)
    #undef DOC_CONNECT

    // Eventually pack and display everything
    gtk_container_add(GTK_CONTAINER(scrolled), browser->document);
    sokoke_widget_set_visible(browser->menubar, TRUE/*config->bShowMenubar*/);
    sokoke_widget_set_visible(browser->navibar, config->bToolbarNavigation);
    sokoke_widget_set_visible(browser->navibar_search, config->bToolbarSearchbox);
    sokoke_widget_set_visible(browser->bookmarkbar, config->bToolbarBookmarks);
    sokoke_widget_set_visible(browser->statusbar, config->bToolbarStatus);
    gtk_widget_show(GTK_WIDGET(scrolled));
    gtk_widget_show(GTK_WIDGET(browser->document));
    gtk_notebook_set_current_page(GTK_NOTEBOOK(browser->documents), page);

    update_browser_actions(browser);
    gtk_widget_show(browser->window);
    gtk_widget_grab_focus(GTK_WIDGET(browser->navibar_location));

    return browser;
}

// -- configuration related

static CConfig* config_load(const gchar* sFilename, GError* *error)
{
    DEBUG_FUNC
    CConfig* config = g_new0(CConfig, 1);

    GKeyFile* keyFile = g_key_file_new();
    g_key_file_load_from_file(keyFile, sFilename
     , G_KEY_FILE_KEEP_COMMENTS, error);
    #define GET_INT(var, key, default) \
     var = sokoke_key_file_get_integer_default( \
     keyFile, "browser", key, default, NULL)
    #define GET_STR(var, key, default) \
     var = sokoke_key_file_get_string_default( \
     keyFile, "browser", key, default, NULL)
    GET_INT(config->iStartup, "Startup", 1);
    GET_STR(config->sHomepage, "Homepage", "http://www.google.com");
    GET_STR(config->sLocationSearch, "LocationSearch", "http://www.google.com/search?q=%s");
    GET_INT(config->bToolbarNavigation, "ToolbarNavigation", TRUE);
    GET_INT(config->bToolbarBookmarks, "ToolbarBookmarks", FALSE);
    GET_INT(config->bToolbarStatus, "ToolbarStatus", TRUE);
    //GET_INT(config->bToolbarDownloads, "ToolbarDownloads", FALSE);
    GET_INT(config->iToolbarStyle, "ToolbarStyle", 0);
    GET_INT(config->iToolbarIconSize, "ToolbarIconSize", 0);
    GET_INT(config->bToolbarSearchbox, "ToolbarSearchbox", TRUE);
    GET_INT(config->bPanels, "Panels", FALSE);
    GET_INT(config->iPanelActive, "PanelActive", 0);
    GET_INT(config->iPanelPos, "PanelPos", 0);
    GET_STR(config->sPanelPageholder, "PanelPageholder", "http://www.google.com");
    GET_INT(config->iTabSize, "TabSize", 10);
    GET_INT(config->iTabClose, "TabClose", 1);
    GET_INT(config->iPopupRule, "PopupRule", 2);
    #undef GET_INT
    #undef GET_STR

    return config;
}

gboolean config_save(CConfig* config, const gchar* sFilename, GError* *error)
{
    DEBUG_FUNC
    GKeyFile* keyFile = g_key_file_new();

    g_key_file_set_integer(keyFile, "browser", "Startup", config->iStartup);
    g_key_file_set_string(keyFile, "browser", "Homepage", config->sHomepage);
    g_key_file_set_string(keyFile, "browser", "LocationSearch", config->sLocationSearch);
    g_key_file_set_integer(keyFile, "browser", "ToolbarNavigation", config->bToolbarNavigation);
    g_key_file_set_integer(keyFile, "browser", "ToolbarBookmarks", config->bToolbarBookmarks);
    //g_key_file_set_integer(keyFile, "browser", "ToolbarDownloads", config->bToolbarDownloads);
    g_key_file_set_integer(keyFile, "browser", "ToolbarStatus", config->bToolbarStatus);
    g_key_file_set_integer(keyFile, "browser", "ToolbarStyle", config->iToolbarStyle);
    g_key_file_set_integer(keyFile, "browser", "ToolbarIconSize", config->iToolbarIconSize);
    g_key_file_set_integer(keyFile, "browser", "ToolbarSearchbox", config->bToolbarSearchbox);
    g_key_file_set_integer(keyFile, "browser", "Panels", config->bPanels);
    g_key_file_set_integer(keyFile, "browser", "PanelActive", config->iPanelActive);
    g_key_file_set_integer(keyFile, "browser", "PanelPos", config->iPanelPos);
    g_key_file_set_string(keyFile, "browser", "PanelPageholder", config->sPanelPageholder);
    g_key_file_set_integer(keyFile, "browser", "TabSize", config->iTabSize);
    g_key_file_set_integer(keyFile, "browser", "TabClose", config->iTabClose);
    g_key_file_set_integer(keyFile, "browser", "PopupRule", config->iPopupRule);

    //g_key_file_set_string(keyFile, "content", "LoadImages", config->bLoadImages);
    //g_key_file_set_string(keyFile, "content", "PrintBackgrounds", config->bPrintBackgrounds);
    //g_key_file_set_string(keyFile, "content", "JavaScript", config->bJavaScript);
    //g_key_file_set_string(keyFile, "content", "StandardFont", config->sStandardFont);
    //g_key_file_set_string(keyFile, "content", "SerifFont", config->sSerifFont);
    //g_key_file_set_string(keyFile, "content", "SansFont", config->sSansFontFont);
    //g_key_file_set_string(keyFile, "content", "FixedFont", config->sFixedFont);
    //g_key_file_set_string(keyFile, "content", "MinFontSize", config->iMinFontSize);

    gboolean bSaved = sokoke_key_file_save_to_file(keyFile, sFilename, error);
    g_key_file_free(keyFile);

    return bSaved;
}

// -- stock icons

void init_stock_icons(void)
{
    // Based on ephy_stock_icons_init from lib/ephy-stock-icons.c from epiphany.
    // Copyright (C) 2002 Jorn Baayen
    GtkIconFactory *factory;
    GtkIconSet *icon_set;
    GtkIconSource *icon_source;
    int i;
    static const gchar* icon_theme_items[] =
    {
       STOCK_LOCK_OPEN
     , STOCK_LOCK_SECURE
     , STOCK_LOCK_BROKEN
     , STOCK_SCRIPT
     , STOCK_THEME
     , STOCK_USER_TRASH
    };
    static const GtkStockItem items[] =
    {
       { STOCK_BOOKMARK,         "Bookmark", 0, 0, NULL }
     , { STOCK_BOOKMARK_NEW,     "New bookmark", 0, 0, NULL }
     , { STOCK_BOOKMARKS,        "_Bookmarks", 0, 0, NULL }
     , { STOCK_DOWNLOADS,        "_Downloads", 0, 0, NULL }
     , { STOCK_CONSOLE,          "_Console", 0, 0, NULL }
     , { STOCK_EXTENSIONS,       "_Extensions", 0, 0, NULL }
     , { STOCK_FORM_FILL,        "_Form fill", 0, 0, NULL }
     , { STOCK_HISTORY,          "History", 0, 0, NULL }
     , { STOCK_HOMEPAGE,         "Homepage", 0, 0, NULL }
     , { STOCK_LOCATION,         "Location entry", 0, 0, NULL }
     , { STOCK_NEWSFEED,         "Newsfeed", 0, 0, NULL }
     , { STOCK_PLUGINS,          "_Plugins", 0, 0, NULL }
     , { STOCK_POPUPS_BLOCKED,   "Blocked popups", 0, 0, NULL }
     , { STOCK_SOURCE_VIEW,      "View source", 0, 0, NULL }
     , { STOCK_TAB_CLOSE,        "Close tab", 0, 0, NULL }
     , { STOCK_TAB_NEW,          "New tab", 0, 0, NULL }
     , { STOCK_WINDOW_CLOSE,     "Close window", 0, 0, NULL }
     , { STOCK_WINDOW_NEW,       "New window", 0, 0, NULL }
     #if !GTK_CHECK_VERSION(2, 10, 0)
     , { GTK_STOCK_SELECT_ALL,   "_Select all", 0, 0, NULL }
     #endif
    };
    // TODO: Use fallback icons with _ suffix
    factory = gtk_icon_factory_new ();
    for(i = 0; i < (int)G_N_ELEMENTS(items); i++)
    {
     icon_source = gtk_icon_source_new();
     gtk_icon_source_set_icon_name (icon_source, items[i].stock_id);
     icon_set = gtk_icon_set_new();
     gtk_icon_set_add_source(icon_set, icon_source);
     gtk_icon_source_free(icon_source);
     gtk_icon_factory_add(factory, items[i].stock_id, icon_set);
     gtk_icon_set_unref(icon_set);
    }
    gtk_stock_add_static(items, G_N_ELEMENTS(items));

    for(i = 0; i < (int)G_N_ELEMENTS(icon_theme_items); i++)
    {
     icon_source = gtk_icon_source_new();
     gtk_icon_source_set_icon_name(icon_source, icon_theme_items[i]);
     icon_set = gtk_icon_set_new();
     gtk_icon_set_add_source(icon_set, icon_source);
     gtk_icon_source_free(icon_source);
     gtk_icon_factory_add(factory, icon_theme_items[i], icon_set);
     gtk_icon_set_unref(icon_set);
    }
    gtk_icon_factory_add_default(factory);
    g_object_unref(factory);

    // TODO: multi-head?
    // TODO: What about this?
    /*gtk_icon_theme_append_search_path(gtk_icon_theme_get_default()
     , SHARE_DIR G_DIR_SEPARATOR_S "icons");*/
}

// -- main function

int main(int argc, char* *argv)
{
    DEBUG_FUNC

    INIT_CRASHSAFE
    g_set_application_name(PACKAGE_NAME);

    // Parse cli options
    gint repeats = 2;
    gboolean bVersion = FALSE;
    GOptionEntry entries[] =
    {
     { "repeats", 'r', 0, G_OPTION_ARG_INT, &repeats, "An unused value", "N" },
     { "version", 'v', 0, G_OPTION_ARG_NONE, &bVersion, "Display program version", NULL }
    };

    GError* error = NULL;
    if(!gtk_init_with_args(&argc, &argv, "[URI]", entries
     , NULL/*GETTEXT_PACKAGE*/, &error))
    {
     g_error_free(error);
     return 1;
    }

    if(bVersion)
    {
     g_print(PACKAGE_STRING " - Copyright (c) 2007 Christian Dywan\n\n"
             "GTK+2:      " GTK_VER "\n"
	     "WebKit:     " WEBKIT_VER "\n"
	     "Libsexy:    " LIBSEXY_VER "\n"
             "GetText:    N/A\n"
	     "\n"
	     "CrashSafe:  " SOKOKE_CRASHSAFE_ "\n"
	     "Debugging:  " SOKOKE_DEBUG_ "\n"
	     "Engine:     " MIDORI_ENGINE_ "\n"
             "\n"
             "Please report comments, suggestions and bugs to:\n"
             "\t" PACKAGE_BUGREPORT "\n"
             "Check for new versions at:\n"
             "\thttp://software.twotoasts.de\n");
     return 0;
    }

    // Load configuration files
    // TODO: What about default config in a global config folder?
    gchar* sConfig = g_build_filename(g_get_user_config_dir(), PACKAGE_NAME, NULL);
    g_mkdir_with_parents(sConfig, 0755);
    gchar* sConfigFile = g_build_filename(sConfig, "config", NULL);
    config = config_load(sConfigFile, &error); g_free(sConfigFile);
    if(error != NULL)
    {
     g_print("Notice: No configuration loaded. %s", error->message);
     g_error_free(error);
    }
    gchar* sAccelsFile = g_build_filename(sConfig, "accels", NULL);
    gtk_accel_map_load(sAccelsFile); g_free(sAccelsFile);
    /*gchar* sBookmarkFile = g_build_filename(sConfig, "bookmarks.xbel", NULL);
    bookmarks = g_bookmark_file_new();
    if(!g_bookmark_file_load_from_file(bookmarks, sBookmarkFile, &error))
     g_warning("Bookmarks couldn't be loaded. %s", error->message); g_free(sBookmarkFile);*/
    g_free(sConfig);

    //browser_list = g_list_alloc();
    accel_group = gtk_accel_group_new();

    init_stock_icons();
    #ifdef MIDORI_WEBKIT
    webkit_gtk_init();
    #endif

    // TODO: Handle any number of separate uris from argv
    gchar* sHomepage = argc > 1 ? argv[1]
     : (config->iStartup == 1 ? config->sHomepage : "http://about:blank");
    // Open as many tabs as we have uris, seperated by pipes
    CBrowser* browser = NULL;
    gchar* sUri = strtok(g_strdup(sHomepage), "|");
    while(sUri != NULL)
    {
     browser = browser_new(browser);
     /* TODO: Use g_idle_add to prevent code from being called
     before the main loop is running */
     document_load_uri(browser->document, magic_uri(sUri, config->sLocationSearch));
     sUri = strtok(NULL, "|");
    } g_free(sUri);

    gtk_main();
    DEBUG_FUNC

    // Save configuration files
    sConfig = g_build_filename(g_get_user_config_dir(), PACKAGE_NAME, NULL);
    g_mkdir_with_parents(sConfig, 0755);
    sConfigFile = g_build_filename(sConfig, "config", NULL);
    if(!config_save(config, sConfigFile, &error))
    {
     g_warning("Configuration couldn't be saved. %s", error->message);
     g_error_free(error);
    }
    g_free(sConfigFile);
    sAccelsFile = g_build_filename(sConfig, "accels", NULL);
    gtk_accel_map_save(sAccelsFile); g_free(sAccelsFile);
    /*sBookmarkFile = g_build_filename(sConfig, "bookmarks.xbel", NULL);
    if(!g_bookmark_file_to_file(bookmarks, sBookmarkFile, NULL))
     g_warning("Bookmarks couldn't be saved."); g_free(sBookmarkFile);*/
    g_free(sConfig);
    return 0;
}
